;Spc2Midi - DSP Miscellaneous
;Routines related to the DSP, but directly to emulation of it.
;
; DumpSoundBuffer
; DumpSoundBuffer.Info
; SeekSongPos
; SeekSongPos.Init
; SeekSongPos.Copy
; NewSample.NextFreeIdx
; NewSample.GetNumberedNames
; NewSample.NextNumberedName
; NewSample.CreateName
; MatchBrrSamples
;
;=============================================================================

section .text
;
; Seeks to any point in the buffer, given a tick time count, updating the
; DSP registers and returning which voices were turned on or off. Mainly used
; to fast forward or rewind.
;
; SetKeys can then update the seek structure with any changes to their
; pitches, volumes, or instruments.
;
; To fit more in the dsp buffer, only the time information of
; important dsp writes is stored. The simple packing method uses the top
; byte to flag whether the dword is a timer value or register/value write.
; Since no song should ever exceed 18 hours, a value of 0FFh can safely
; indicate it is a write. Anything less than that should be considered
; time information.
;
; (edi=song pos struct, edx=new time in ticks)
; (bl=voices on, bh=voices off; edi)
SeekSongPos:
    ; if new time >= old time
    ;   seek forward
    ; elif time = 0
    ;   clear all variables in structure
    ; else new time < old time
    ;   seek backward
    ; endif

    xor ebx,ebx                 ;zero keys on & off
    mov esi,[edi+SongPos.Pos]
    xor ecx,ecx                 ;zero extend
    cmp [edi+SongPos.Time],edx
    je near .SetVoices
    mov [edi+SongPos.Time],edx
    jbe .Forward
    test edx,edx
    jle .Init
    jmp .Reverse

;
.Init:
    push edi
    mov esi,FileDspRam
    cld
    mov ecx,128/4
    rep movsd
    xor eax,eax
    %if SongPos.DspRegs != 0
    %error "DspEmu.Init: DspRegs must come first in structure def."
    %endif
    mov ecx,(SongPos.SizeOf-128)/4
    rep stosd
    pop edi
    ;mov bl,[edi+SongPos.DspRegs+DspReg.KeyOn]
    mov ebx,0FF00h              ;turn off all, turn on none
    mov al,[MutedVoices]
    mov [edi+SongPos.Muted],al
    jmp short .SetVoices

;
; (edi=seek structure, esi=buffer offset, edx=time, ebx&ecx=0)
    ; start from previous position
    ; loop until buffer pos >= buffer end or note time >= specified time
    ;   if note on
    ;     flag note was turned on
    ;   elif note off
    ;     flag note was turned off
    ;   else
    ;     set dsp register
    ;   endif
    ; endloop
    ; set instrument and pitch for all voices turned on

.Next:
    add esi,byte 4
.Forward:
    cmp [DspBufferEnd],esi      ;beyond end of buffer?
    jbe .EndLoop

    ; read next time value or register write
    mov eax,[DspBuffer+esi]
    cmp eax,0FF000000h          ;time value or register write?
    jb .CheckTime

    cmp al,DspReg.KeyOn
    je .Kon
    cmp al,DspReg.KeyOff
    je .Kof
    mov cl,al
    mov [edi+SongPos.DspRegs+ecx],ah
    jmp short .Next
.Kon:
    or bl,ah                    ;keys on |= new keys on
    ;not ah
    ;and bh,ah                  ;keys off &= !new keys on
    jmp short .Next
.Kof:
    or bh,ah                    ;keys off |= new keys off
    not ah
    and bl,ah                   ;keys on &= !new keys off
    jmp short .Next
.CheckTime:
    cmp eax,edx                 ;loop until >= desired time
    jb .Next
    ;jmp short .EndLoop (fall through)

;
.EndLoop:
    mov [edi+SongPos.Pos],esi   ;store buffer position

.SetVoices:
    ; set flags for any keys turned on, turned off, or un/muted by the user
    mov al,[edi+SongPos.Muted]
    and bh,[edi+SongPos.Active] ;voices off &= active voices
    xor al,[MutedVoices]
    xor [edi+SongPos.Active],bh ;turn off old voices
    xor [edi+SongPos.Muted],al
    or  [edi+SongPos.Active],bl ;turn on new voices
    mov [edi+SongPos.Kon],bx
    or al,bl
    or al,bh
    mov [edi+SongPos.Changed],al
    ret

;
; Called immediately after seeking, it calculates the pitch and volume for
; each voice. This only needs to be called by routines that need such
; conversion, like the note display and FM synthesis. The wave routine does
; this at all.
;
; (edi=seek structure, bl=keys on)
.SetNewKeys:
    mov bl,[edi+SongPos.Kon]
    jmp short .SetKeys
.SetAllKeys:
    mov bl,[edi+SongPos.Active]
.SetKeys:
    ; calculate pitches for new notes turned on
    ; don't care about notes that were turned off
    push ebp
    xor ecx,ecx                 ;start with voice 0
    mov esi,edi
    jmp short .KonStart
.KonSet:
    ; MIDI note = HertzNoteTbl[ (DspPitch * Frequency) >> 12 ]

    movzx edx,byte [esi+SongPos.DspRegs+DspReg.SourceNumber] ;get sample number
    movzx eax,word [esi+SongPos.DspRegs+DspReg.PitchLow]
    mov edx,[BrrList.SampleIdx+edx*4]
    imul eax,[SampleList.Freq+edx*4]
    shr eax,11

    ; map frequency (play rate * apparent pitch) to MIDI note
    cmp eax,12543*2             ;> highest MIDI note's frequency (G10)
    jbe .PitchOk
    mov al,127                  ;limit to highest MIDI note
    jmp short .PitchSet
.PitchOk:
    mov al,[HertzNoteTbl+eax]
.PitchSet:
    mov [edi+SongPos.Pitch+ecx],al

    ; set GM instrument for sample
    mov dl,[SampleList.Instrument+edx*4]
    mov [edi+SongPos.Instrument+ecx],dl

    ; get combined volume of left & right channels
    movsx ebp,byte [esi+DspReg.VolLeft]
    movsx edx,byte [esi+DspReg.VolRight]
    mov eax,ebp
    xor ebp,edx
    jns .SameVolSign            ;left and right volume have same sign
    neg edx                     ;make right volume same sign as left
.SameVolSign:
    add edx,eax                 ;left + right
    jns .VolPositive
    neg edx                     ;make volume value positive
.VolPositive:
    mov [edi+SongPos.Volume+ecx*4],edx

.KonNext:
    inc ecx                     ;next voice
    add esi,byte 16             ;next register set
.KonStart:
    shr bl,1
    jc .KonSet
    jnz .KonNext
    pop ebp
    ret

;
; (esi=src song pos struct, edi=dest ssong pos struct)
; (; esi)
.Copy:
    cld
    mov ecx,SongPos.SizeOf/4
    rep movsd
    sub esi,SongPos.SizeOf
    ret

;
; (edi=seek structure, esi=buffer offset, edx=time, ebx&ecx=0)
.Reverse:
    push esi
.RevValidTime:
    mov ecx,esi
.RevScanNext:
    sub esi,byte 4
    jl .RevScanEnd
    mov eax,[DspBuffer+esi]
    cmp eax,0FF000000h          ;time value or register write?
    jae .RevScanNext
    cmp eax,edx                 ;loop until >= desired time
    jae .RevValidTime
.RevScanEnd:
    pop esi

    xor edx,edx
.RevNext:
    cmp esi,ecx
    jbe near .EndLoop
    sub esi,byte 4
    mov eax,[DspBuffer+esi]
    cmp eax,0FF000000h          ;time value or register write?
    jb .RevNext
    cmp al,DspReg.KeyOn         ;note that reverse seek also flips on->off
    je .RevKof
    cmp al,DspReg.KeyOff        ;..and off->on
    je .RevKon
    mov dl,al
    shr eax,8
    mov [edi+SongPos.DspRegs+edx],ah
    jmp short .RevNext
.RevKon:
    or bl,ah                    ;keys on |= new keys on
    ;not ah
    ;and bh,ah                   ;keys off &= !new keys on
    jmp short .RevNext
.RevKof:
    or bh,ah                    ;keys off |= new keys off
    not ah
    and bl,ah                   ;keys on &= !new keys off
    jmp short .RevNext


;
; Following are a set of functions to help with adding new samples.
NewSample:

;
; Returns and allocates the next available sample index. Does not initialize
; the variables for that sample!
;
; ()
; (cf=error-no more samples, eax=sample index; !)
.NextFreeIdx:
    xor esi,esi
    mov ecx,SampleList.Total/32 ;(8 bits per byte, 4 bytes per dword)
.NextSample:
    mov edx,[SampleList.Used+esi]
    not edx                     ;invert bits, looking for free ones, not used
    bsf eax,edx
    jnz .IdxFound
    add esi,byte 4
    loop .NextSample
    stc
    ret
.IdxFound:
    lea eax,[esi*8+eax]
    inc dword [SampleList.Entries]
    cmp [SampleList.LastEntry],eax
    jae .LastEntrySame
    mov [SampleList.LastEntry],eax
.LastEntrySame:
    bts [SampleList.Used],eax
    ;clc
    ret

;
; It's best to name samples according to the instrument they most sound like
; (of course there are plenty of sounds in video games that have absolutely
; no musical equivalent). What do you do though when there are multiple
; variations of the same instrument? Maybe various types type of flute? If
; you know instruments well, you could name them after their real-world types
; (Pan Flute, Wood Flute); but if you're like me, and know music well but not
; necessarily the whole realm of music, it would be easier to simply number
; each instance. This routine helps with that numbering by finding all
; occurences of a specific instrument so that a new number can be returned.
;
; For example, if "Flute", "Flute 2", and "Flute  4" are existing names, the
; next available one is "Flute 3".
;
; Before searching for a free number, it strips the given name of any number
; it may have following.
;   "Flute 2"          >> "Flute ?"
;   "Flute-2"          >> "Flute-?"
;   "Flute #2"         >> "Flute #?"
;   "Flute2"           >> "Flute?"
;   "Flute3 4"         >> "Flute3 ?"
;   "Flute Bright 3"   >> "Flute Bright ?"
;   "Flute Bright   3" >> "Flute Bright ?"
;   "Flute Bright A"   >> "Flute Bright A ?"
;   "23"               >> "?"
;   ""                 >> "?"
;
; When comparing the search name with other names, it is not case sensitive
; but only chooses names that completely match the text portion.
;   "Wood Flute"      yes
;   "Wood Flute    "  yes
;   "Wood Flute 2"    yes
;   "Wood Flute   56" yes
;   "wOoD fLuTe 3"    yes
;   "Wood Flute A"    no
;   "Wood Glock 2"    no
;
; (esi=sample name blen string ptr)
; (!)
.GetNumberedNames:
    ; copy sample name and lowercase second copy
    ; strip sample name of trailing numbers or space
    ; find all samples of same name and record their number in table
    cld
    mov ecx,.MaxFlags/32        ;(32 = 8 bits per byte, 4 bytes per dword)
    mov eax,-1                  ;clear table, filling with all ones
    mov edi,.NameFlags          ;to indicate names are free
    rep stosd
    mov edi,.Name
    mov ecx,SampleList.NameSize/4
    rep movsd
    mov esi,.Name
    mov edi,.NameLcase
    mov ecx,SampleList.NameSize-2
    movsb                       ;transfer length byte
.NextUpperCase:
    lodsb
    cmp al,'A'
    jb .NotUpperCase
    cmp al,'Z'
    ja .NotUpperCase
    or al,32  ;make lowercase
.NotUpperCase:
    stosb
    dec ecx
    jg .NextUpperCase

;
    ; get name length
    movzx ecx,byte [.Name]
    test ecx,ecx
    jz .SetNameMinLen

    ; ignore any trailing numbers
    mov edx,ecx
.SkipDigits:
    mov al,[.Name+ecx]
    cmp al,'9'
    ja .NoSpace
    cmp al,' '
    je .SkipSpaces
    cmp al,'0'
    jb .NoSpace
    dec ecx
    jg .SkipDigits
    jmp short .SetNameMinLen

    ; if offset = name length
    ;   append space to name
    ; else offset < name length
    ;   assume number is appended directly to end of word
    ; endif
.NoSpace:                       ;number is joined with name
    cmp ecx,edx
    jb .SetNameMinLen
    jmp short .AppendSpace

    ; trim any spaces
    ; loop while offset > 0 && char[offset++] = ' '
.SkipSpaces:
    dec ecx
    jle .SetNameMinLen
    cmp byte [.Name+ecx],' '
    je .SkipSpaces

.AppendSpace:
    cmp ecx,SampleList.NameSize-2
    jae .SetNameMinLen
    mov [.NameMinLen],ecx
    inc ecx
    mov byte [.Name+ecx],' '
    jmp short .SetNameLen

.SetNameMinLen:
    mov [.NameMinLen],ecx
.SetNameLen:
    mov edx,SampleList.NameSize-2
    mov [.NameLen],ecx
    sub edx,ecx
    mov [.NumLen],edx

;
    ; find all similar sample names and record their numbers in table
    mov edx,SampleList.Name
    xor ebx,ebx
.SampleNameNext:
    bt dword [SampleList.Used],ebx  ;ensure sample info isn't null junk
    jnc .Mismatch

    ; if compared sample name >= original sample name
    ;   loop while > 0
    ;     if char = 'A'-'Z' make lowercase
    ;     if mismatch break loop
    ;     char ptr++
    ;   endloop
    ;   ...
    mov ecx,[.NameMinLen]
    cmp [edx],cl
    jb .Mismatch
    lea esi,[edx+1]             ;first character after len byte
    mov edi,.Name+1
    jmp short .CompareStart
.CompareNext:
    lodsb
    cmp al,'Z'
    ja .CompareChar
    cmp al,'A'
    jb .CompareChar
    or al,32  ;make lowercase
.CompareChar:
    cmp [edi],al
    jne .Mismatch
.CompareStart:
    dec ecx
    jns .CompareNext

.NameMatch:
    ;   get number from name and clear flag in table
    ;   validation is intermixed with the number conversion
    ;   loop while char is valid
    ;     skip any spaces
    ;     number = number * 10 + char
    ;     char ptr++
    ;   endloop
    ;   if last char was null && flag number < max flags
    ;     clear flag at number
    ;   end if
    xor eax,eax
    xor ecx,ecx
    jmp short .NumStart
.NumAdd:
    imul ecx,10
    add ecx,eax
.NumNext:
    inc esi
.NumStart:
    mov al,[esi]
    sub al,'0'
    cmp al,9
    jbe .NumAdd
    cmp al,' '-'0'
    je .NumNext
    cmp al,-'0'
    jne .Mismatch
    cmp ecx,.MaxFlags
    jae .Mismatch
    btr dword [.NameFlags],ecx

.Mismatch:
    ; endloop while < total sample attribute entries
    inc ebx
    add edx,byte SampleList.NameSize
    cmp [SampleList.LastEntry],ebx
    jae near .SampleNameNext

    ret

;
; ()
; (eax=next free sample number; edi)
.NextNumberedName:
    ; search through all flags for a free numbered name
    xor esi,esi
    mov ecx,.MaxFlags/32 ;(8 bits per byte, 4 bytes per dword)
.NextName:
    mov edx,[.NameFlags+esi]
    bsf eax,edx
    jnz .NameFound
    add esi,byte 4
    loop .NextName
    mov eax,.MaxFlags-1
    jmp short .NoNameFound
.NameFound:
    lea eax,[esi*8+eax]
    btr [.NameFlags],eax
.NoNameFound:
    ret

;
; Based on the name passed to NewSampleName, along with a given number, this
; will build the string name.
;
; (edi=blen string dest, eax=sample number)
; ()
.CreateName:
    push edi
    mov ecx,[.NameLen]
    mov esi,.Name+1
    inc edi
    rep movsb

    push edi
    call NumToString
    lea esi,[NumToString.Buffer+ecx]
    neg ecx
    pop edi
    add ecx,byte NumToString.DefMaxLen
    cld
    cmp [.NumLen],ecx
    jae .NumLenOk
    mov ecx,[.NumLen]
.NumLenOk:
    rep movsb
    xor eax,eax
    mov ecx,edi
    stosb

    pop edi
    sub ecx,edi
    dec ecx
    mov [edi],cl
    ret


section .bss
alignb 4
.MaxFlags           equ SampleList.Total
.NameFlags          resb (.MaxFlags+7)/8 ;bit flags telling name free or not
.NameLen            resd 1      ;significant length of sample name
.NameMinLen         resd 1      ;minimum length of sample name
.NumLen             resd 1      ;significant length of sample name
.Name               resb SampleList.NameSize  ;last used name
.NameLcase          resb SampleList.NameSize  ;last used name
section .text

;
; Matches BRR samples in the local SPC source directory to any existing
; samples already defined in the global list. New BRR samples (no match found
; in the global list) are added to it with default attributes.
;
; Matching samples is necessary because of how often games have entries in
; their source directory that point to the exact same sound. Gs Mikami is
; good example of this. The first four are all the same sample. Then the next
; four all point to another sample. Without this matching, you would get
; eight samples from above, when really there were only two.
;
; This function is called upon after loading an SPC state. A subfunction of
; it is called after a key on for any new samples.
;
; () ()
MatchBrrSamples:
    mov esi,Text.IdentifyingSamples
    call StatusMessage

; match all samples in the global list to any duplicates in the local list
    mov eax,1                   ;start at sample 1 (skipping default)
.NextSample:
    bt dword [SampleList.Used],eax  ;ensure sample info isn't null junk
    jnc .NullSample
    mov ebx,[SampleList.Checksum+eax*4]
    mov ecx,[SampleList.Length+eax*4]
    mov edx,[SampleList.LoopLen+eax*4]
    mov dword [SampleList.BrrIdx+eax*4],-1

    ; count negative (255-0) so that the lowest BRR number will be
    ; pointed to in the sample's BRR index
    mov esi,255
.NextBrr:
    cmp [BrrList.Checksum+esi*4],ebx
    jne .Mismatch
    cmp [BrrList.Length+esi*4],ecx
    jne .Mismatch
    cmp [BrrList.LoopLen+esi*4],edx
    jne .Mismatch

    ; BRR sample matched sample from list
    ; make both point to each other, sample to BRR, BRR to sample
    mov [SampleList.BrrIdx+eax*4],esi
    mov [BrrList.SampleIdx+esi*4],eax
    ; if a BRR sample matches one from the global sample list, it must be
    ; valid, so mark it accordingly
    bts [BrrList.Valid],esi

.Mismatch:
    dec esi
    jns .NextBrr

.NullSample:
    inc eax
    cmp [SampleList.LastEntry],eax
    jae .NextSample

;
; Create names and default attributes for any new samples.
;
; Called after the key on of a new sample.
; Loading a new SPC also falls through to here.
.New:
    mov esi,.DefaultName
    call NewSample.GetNumberedNames

    xor esi,esi
.NextNewBrr:
    bt [BrrList.Valid],esi
    jnc near .SkipSample
    cmp dword [BrrList.SampleIdx+esi*4],0
    jg near .SkipSample

    push esi
    call NewSample.NextFreeIdx
    pop esi
    jc near .NoMoreSamples
    ;(eax=new index)

    ; copy BRR attributes to sample list and initialize other sample attrs
    mov ebx,[BrrList.Checksum+esi*4]
    mov ecx,[BrrList.Length+esi*4]
    mov edx,[BrrList.LoopLen+esi*4]
    push dword [DefaultGmPatch]
    mov [SampleList.BrrIdx+eax*4],esi
    mov [SampleList.Checksum+eax*4],ebx
    mov [SampleList.Length+eax*4],ecx
    mov [SampleList.LoopLen+eax*4],edx
    ;mov [SampleList.Instrument+eax*4],dword DefaultSample.Instrument
    pop dword [SampleList.Instrument+eax*4]
    mov [SampleList.SoundFx+eax*4],dword DefaultSample.SoundFx
    mov [SampleList.Volume+eax],byte DefaultSample.Volume
    mov [SampleList.Interpolation+eax],byte DefaultSample.Interpolation

%if 0
    ; This was an attempt to automatically tune samples according to their
    ; loop length. It worked for a few songs, but made the majority of songs
    ; sound worse.
    push ecx
    push edx
    push eax
    cmp edx,16
    mov ecx,DefaultSample.Freq
    jb .DefaultHz
    mov ecx,edx
.KeepMainLen:
    test ecx,ecx
    jg .SampleLenNz
    mov ecx,16
.SampleLenNz:
    xor edx,edx
    mov eax,32000<<17
    div ecx
    mov ecx,17
.NextHigherHz:
    shr eax,1
    cmp eax,660
    jb .HertzOk
    shr eax,1
    dec ecx
    jg .NextHigherHz
.HertzOk:
    mov ecx,eax
.DefaultHz:
    pop eax
    mov [SampleList.Freq+eax*4],ecx
    pop edx
    pop ecx
%else
    mov [SampleList.Freq+eax*4],dword DefaultSample.Freq
%endif

    push esi
    jmp short .StartDupCheck
.NextCheckDupBrr:
    cmp [BrrList.Checksum+esi*4],ebx
    jne .NotDup
    cmp [BrrList.Length+esi*4],ecx
    jne .NotDup
    cmp [BrrList.LoopLen+esi*4],edx
    jne .NotDup
.StartDupCheck:
    ; another BRR sample matched sample from list
    mov [BrrList.SampleIdx+esi*4],eax
.NotDup:
    inc esi
    cmp esi,256
    jb .NextCheckDupBrr
    pop esi

    mov edi,eax                 ;get sample list index for name ptr
    shl edi,SampleList.NameShl
    mov eax,esi                 ;copy BRR sample number for name
    add edi,SampleList.Name
    bt [NewSample.NameFlags],esi
    push esi
    jc .CreateName
    call NewSample.NextNumberedName
.CreateName:
    call NewSample.CreateName
    pop esi

.SkipSample:
    inc esi
    cmp esi,256
    jb near .NextNewBrr
.NoMoreSamples:

    ret

section .data
.DefaultName:   db 4,"wave",0
section .text


;
; Exports cached sample buffer to wave file.
;
; () ()
DumpSoundBuffer:
    ; clear sample buffer and decompress all valid samples
    cld
    mov edi,SamplesBuffer
    mov ecx,SamplesBufferSize/4
    xor eax,eax
    rep stosd
    call DecompressAllSamples

    ; create file and export wave
    mov ah,3Ch                  ;DOS: Create file
    xor ecx,ecx                 ;no attributes
    mov edx,SoundDumpFileName
    int 21h
    jc .FileError
    mov ebx,eax
    mov edx,SoundDumpHeader
    mov ecx,SoundDumpHeader.Size
    mov ah,40h                  ;DOS: Write to file
    int 21h
    mov edx,SamplesBuffer
    mov ecx,SamplesBufferSize
    mov ah,40h                  ;DOS: Write to file
    int 21h
    mov ah,3Eh                  ;DOS: Close file
    int 21h
.FileError:
    ret

;
; Writes info about each sample to screen (can be redirected to file)
;
; () ()
.Info:
    mov edx,Text.SampleInfoTitles
    call WriteString

    xor ecx,ecx
.NextSampleInfoLine:
    push ecx

    mov eax,'    '
    bt [BrrList.Used],ecx
    jnc .Unused
    mov ah,'u'
.Unused:
    bt [BrrList.Valid],ecx
    jnc .Invalid
    mov al,'v'
.Invalid:
    mov [Text.SampleInfo+20],ax

    push dword [BrrList.SampleIdx+ecx*4]
    push dword [BrrList.Checksum+ecx*4]
    push dword [BrrList.LoopLen+ecx*4]
    push dword [BrrList.Length+ecx*4]
    push dword [BrrList.Offset+ecx*4]

    mov eax,ecx                 ;sample number
    mov edi,Text.SampleInfo
    mov ecx,3
    call NumToString.UsingDLen

    movzx eax,word [esp]        ;sample offset
    mov edi,Text.SampleInfo+7
    mov ecx,5
    call NumToString.UsingDLen

    pop eax                     ;loop offset
    shr eax,16
    mov edi,Text.SampleInfo+13
    mov ecx,5
    call NumToString.UsingDLen

    pop eax                     ;sample length
    mov edi,Text.SampleInfo+24
    mov ecx,5
    call NumToString.UsingDLen

    pop eax                     ;loop length
    mov edi,Text.SampleInfo+30
    mov ecx,5
    call NumToString.UsingDLen

    pop eax                     ;checksum
    mov edi,Text.SampleInfo+37
    mov ecx,32                  ;bits
    call MakeHexNum

    pop esi
    cld
    mov edx,30
    mov edi,Text.SampleInfo+47
    shl esi,SampleList.NameShl
    jle .NoSampleName
    add esi,SampleList.Name
    movzx ecx,byte [esi]
    inc esi
    sub edx,ecx
    ja .NameLenOk
    add ecx,edx
    xor edx,edx
.NameLenOk:
    rep movsb
.NoSampleName:
    mov al,' '
    mov ecx,edx
    rep stosb

    mov edx,Text.SampleInfo
    call WriteString
    pop ecx
    inc cl
    jnz near .NextSampleInfoLine
    ret


;
%if 0 ;old unused function
OutputDspWrites:

    mov ah,3Ch                  ;create file
    xor ecx,ecx                 ;no attributes
    mov edx,DspFileName
    int 21h
    mov edx,Text.DspOutputError
    jc EndProgram.ErrorMessage  ;end program and print error message
    mov [DspFileHandle],eax
    mov ebx,eax
    mov ecx,128
    mov edx,DspRam
    mov ah,40h                  ;DOS: Write to file
    int 21h

    mov edi,[DspBufferEnd]
    mov esi,[DspBufferWritten]
    mov [DspBufferWritten],edi
    cmp edi,esi
    jz .End
    sub edi,esi
    mov ebx,[DspFileHandle]
    mov ecx,6
.Next:
    mov ah,40h                  ;DOS: Write to file
    lea edx,[DspBuffer+esi]
    int 21h
    add esi,byte 8
    sub edi,byte 8
    ja .Next
.End:

    mov ah,3Eh              ;close file
    mov ebx,[DspFileHandle]
    int 21h

    ret

DspFileName:        db "out.dsp",0

%endif
