; Spc2Midi - Sound card functions (DA/FM/MPU)
;
; All the routines contained in here are just generic routines that could
; theoretically be used in any program. As a matter of fact, I copied these
; all straight from another program into Spc2Midi. Alone, they have no
; ability to make music.
;
; todo: eliminate FM clicking! (example: voice 0 of SD3 "Swivel")
;        hhm, the clicking is less on my newer SB16, the old 8bit is bad
;       try to silence audio clicking when initialized
;       find out why using MPU with wave freeze my computer. MPU and FM
;        alone work fine. me thinks there is something the SB documentation
;        fails to mention.
;
; GetBlasterVars
; InitFmSound
; WriteFmReg
; TimedWait
; SilenceFm
; SetFmPatch
; InitDaSound
; SoundHandler
; ReadSbDsp
; WriteSbDsp
; InitDmaTransfer
; ClearSbBuffer
; ClearMixBuffer
; CopyMixBuffer
; SyncToDma
; InitMpu
; WaitMpu
; WriteMpu
;
; Except for a few variables in the main source, all these routines are
; fairly independant of the rest of the program - probably because I
; originally wrote them in a separate program ;)
;
; Fm Synthesis:
;   Notes are output (on/off) every 1/30 of a second. This seems to be
;   accurate enough even for dense songs. The routine can handle forwards
;   and backwards playing.
;
; Wave Output:
;   Once turned on, it stays on. Rather than bother with turning the speaker
;   on/off or starting/stopping auto-init, the sound is simply always playing.
;   Even when everything is paused, it is still playing, but playing silence
;   (the sound buffer is cleared). This makes the sound code easier and also
;   eliminates some clicking.
;
;   To make things simpler, I decided to make all sound output mono (for
;   stereo, try any one of several other SPC players). That way, the only
;   complication is whether to play 8bit or 16bit.
;
; Progressive buffering
;   Instead of double buffering (with only two halves), it uses 'progressive'
;   buffering, picking up right where it left off the last transfer and
;   wrapping if necessary. This means the buffer size can always be the same,
;   regardless of the sample rate. If for any reason the buffer ptr and the
;   DMA output become unsynchronized, it will compensate for the difference.
;   This helps if for any reason the sound processing lags or if the user
;   switches away using a multi-tasking OS like MS Windows and then back.
;
; Premixing
;   All sound output is first mixed into a temporary buffer using 32bit
;   precision, to increase accuracy, to lessen overflow problems, to merge
;   multiple sounds sources (playing voices and a sample at the same time,
;   plus anything else I might later want to add) and last, greatly simplify
;   things for me! After everything has been mixed that needs to be, the mix
;   buffer is downsampled to either 8bit or 16bit and copied to the sound
;   card's buffer in low memory.
;
; Sound card interrupt
;   Unlike other sound playing programs (almost every one in existance) this
;   one gets its timing solely from the pc timer, not the sound card. As a
;   matter of fact, it completely ignores any interrupts generated from the
;   card. This is done for several reasons. The PIT is more accurate than the
;   sound card sampling rate, letting this program and the real SNES sync
;   better. The two processes (FM synthesis and wave output) are always
;   coordinated; and the program will play realtime even with no sound card.
;   Of course, this brings up other timing issues with the sound card...
;
; Imprecise sound cards
;   It seems (after carefully checking and rechecking my code) that sound
;   card time constants are rather innaccurrate. Setting a direct sampling
;   rate works great and is near perfectly synchronized, but only SB4.0 or
;   better supports that command :( At first I thought my time constant
;   calculation was off, but even using the formula their own documentation
;   specifies, the play rate is still slightly off. Add to that, the same
;   time constant yields slightly different sampling rates on different sound
;   cards. I've read on the net that cheap sound cards using low precision
;   quartz timers are to blame for the inaccuracy.
;
; Compensating for lost time
;   The problem happens when the DMA and sound generation routine become so
;   out of sync that the two cross streams (like shearing on a VGA screen
;   when a game's update and the adapter refresh meet) causing a short but
;   annoying static sound. The solution? If the mixing falls behind DMA,
;   simply generate more sound for a brief time to catch up. If
;   mixing advances too far ahead of DMA, skip some mixing.
;
;=============================================================================

SbAdr:
.Reset          equ 6
.Fm             equ 8
.Read           equ 0Ah
.Write          equ 0Ch
.WriteStatus    equ 0Ch
.ReadStatus     equ 0Eh
.Ack8bit        equ 0Eh
.Ack16bit       equ 0Fh

SbFmAdr:
;.Timer1Count    equ 2   ;incremented every 80 microseconds
;.Timer2Count    equ 3   ;incremented every 320 microseconds
;.TimerControl   equ 4
;.TimerReset     equ 128 ;clears timer expired flags in port 388h
;.Timer1Mask     equ 64
;.Timer2Mask     equ 32
;.Timer1Start    equ 1
;.Timer2Start    equ 2

SbCmd:
.MidiOut        equ 38h
.ExitDma8bit    equ 0DAh
.ExitDma16bit   equ 0D9h
.PauseDma8bit   equ 0D0h
.PauseDma16bit  equ 0D5h
.GetVersion     equ 0E1h

Pic:
.Mode1          equ  20h
.Mode2          equ 0A0h
.Mask1          equ  21h
.Mask2          equ 0A1h

MpuCmd:
;Data port is 330h, Command port is 331h, Status port is 331h read only
.Reset          equ 0FFh
.Uart           equ 3Fh
.MaskOut        equ 64          ;MPU is ready to accept data or command
.MaskIn         equ 128         ;data is available to read
.RespAck        equ 0FEh


;
section .data
align 4
Sound:
.IoPort:        dd 220h         ;220/240..
.FmPort:        dd 388h         ;228/388
.MpuPort:       dd 300h         ;300/330
;.MixerPort:    dd 220h
.CardVersion:   dd 0            ;2/3/4
.Ldma:          dd 4            ;8bit memory channel (sentinel value set to 4 because DMA channel 4 is invalid)
.Hdma:          dd 4            ;16bit memory channel
.DmaChannel:    dd 4            ;last used channel
.BufferPtr:     dd 0            ;address to base of sound buffer
.BufferSel:     dd 0            ;selector to sound buffer
.BufferPos:     dd 0            ;offset within buffer of last wave data write
.BufferSize     equ 16384       ;4k samples * 2 bytes per 16bit sample * 2 for wrap padding
.TransferSize   equ 65534       ;size given to sound card (for fewer pointless interrupts)
.SampleRate:    dd 32000        ;sound card rate (sound cards are innaccurate)
.MixRate:       dd 32000        ;actual mixing rate
.MinSampleRate  equ 5000
.HighSampleRate equ 23000       ;anything above rate this requires high speed mode (unless DSP >= 4)
.MaxSampleRate  equ 64000       ;might be supported someday...
.MinMixRate     equ 2000
.MaxMixRate     equ 200000      ;anything above rate this requires high speed mode (unless DSP >= 4)
.IntSamples:    dd 735          ;samples to mix per interrupt
.MixSamples:    dd 735          ;samples to mix this interrupt (usually same as IntSamples, but will be twice as many if fallen behind)
.PrevLagDif:    dd 0            ;previous lag difference, (DMA-Mixing)-buffer size
.MinIntSamples  equ 100         ;max samples generated per interrupt; fit enough samples for the maximum playrate (64000/30=2133.33)
.MaxIntSamples  equ 2134        ;max samples generated per interrupt; fit enough samples for the maximum playrate (64000/30=2133.33)
.MixBufferSize  equ 2134*2*4    ;max int samples * 4 for dwords * 2 in case we fall behind
.PicMaskPort:   dd 21h          ;PIC mask port for given IRQ 21=1-7, A1=8-15
.DaInt:         dd 0Dh          ;interrupt number of given IRQ (wave audio)
.Irq:           db 5            ;IRQ to clear

SoundFlags:                     ;which sound capabilities should be used
.Da             equ 1
.DaBit          equ 0
.Fm             equ 2           ;believe it or not, some cards completely lack FM synthesis
.FmBit          equ 1
.Mpu            equ 4
.MpuBit         equ 2
;.Stereo        equ 8           ;stereo/mono
.Force8bit      equ 8           ;force 8bit instead (even if 16bit supported)
.Supports16bit  equ 16          ;>=SB4 is capable of 16bit (even if not used)
.Resync         equ 32          ;flags that resync is needed with DMA
.HighSpeed      equ 64          ;card is using high speed mode for given
.HighSpeedBt    equ 6           ; frequency
.NeedHighSpeed  equ 128         ;old sound card requires high speed
;.ClearedBuffer equ 128         ;sb buffer is clear
                dd .Da|.Fm
                ;dd .Da|.Mpu    ;doesn't work!?
                ;dd 0

; The sound buffer is set in low memory (<16Mb) pointed to by BufferPtr.
; The mixing buffer is where multiple sounds are scaled, amplified, and merged
; before being moved to the sound card buffer. All sample calculations
; internally are 32bit, which must be clipped and converted (8bit/16bit).
; This can allow the program to play 8bit sound, but record 16bit audio.
; Physical memory reserved for it is actually twice as large as the buffer
; alone, just in case part of the allocated block crosses a 64k page.

section .bss
alignb 4
MixBuffer:      resb Sound.MixBufferSize

SbFm:
section .bss
.UsedPatches:   resd 9          ;last used instrument for that voice
.UsedFnumbers:  resb 9          ;last used fnumber of a voice
.UsedVolumes:   resb 9          ;last used volume of a voice
section .data
align 4
.VoicesOff:     dd 0            ;fm voices to turn off soon
.PatchRegs:     db 20h,23h      ;multiplier
                db 40h,43h      ;level
                db 60h,63h      ;attack/decay
                db 80h,83h      ;sustain/release
                db 0E0h,0E3h    ;waveform
                db 0C0h         ;feedback
                db 0            ;null (end)
; a table to compensate for the OPL2's weird register configuration
.VoiceTable:    db 0,1,2,8,9,10,16,17,18,24,25,26
align 4
.PatchData:     incbin "patches.dat"
%if 0
.PatchDefault:  db   2h,  1h;2
                db  10h,  0h
                db  99h,074h;f4
                db  88h,0FFh
                db   0h,  0h
                db   0h
%endif

section .text
;
; Reads variables from BLASTER string. Assumes all settings are correct, and
; that the address, IRQ, & DMA are valid. Don't tell the user an annoying
; message like "invalid blaster setting", because as many different
; computers and different configurations are there are out there, nothing
; can be accepted as standard. However, in the unlikely event the specified
; settings are incorrect (which can happen in the wonderful world of Windows),
; prepare to crash! :/ Well, not neccessarily. Some do... Some don't...
GetBlasterVars:
    ;get DOS environment string (db "BLASTER",0)
    ;get parts  Address,Dma,Interrupt,HighDma,Mpu

%if 0 ;print strings
    mov esi,[Program.Env]
    jmp short .FirstVariable
.NextVariable:
    call WriteString
.FirstVariable:
    mov edx,[esi]
    add esi,byte 4
    test edx,edx
    jnz .NextVariable
    ret
%endif

    ; search through environment for BLASTER variable
    cld
    mov ebx,[Program.Env]
    jmp short .FirstVariable
.NextVariable:
    add ebx,byte 4
    cmp dword [esi],'BLAS'
    jne .FirstVariable
    cmp dword [esi+4],'TER='
    je .VariableFound
.FirstVariable:
    mov esi,[ebx]
    test esi,esi
    jnz .NextVariable
    mov esi,Text.NoBlasterVars
    call StatusMessage
    and byte [SoundFlags],~SoundFlags.Da
    stc
    ret

.VariableFound:
    add esi,byte 8              ;skip "BLASTER=" variable name

.NextElement:
    mov bl,[esi]
    test bl,bl
    je near .End
    inc esi
    cmp bl,32
    je .NextElement

;Converts hex number string in environment variable to number
;(esi=ptr to hex number string) (bl=letter)
.GetNum:
    xor eax,eax             ;set return value to zero
.NextDigit:
    mov dl,[esi]            ;get digit
    sub dl,48               ;"0"=48->0  "9"=57->9
    jc .LastDigit           ;character is less than '0'
    cmp dl,10
    jb .AddPlaceValue       ;digit is 0-9
    and dl,~32              ;make uppercase by turning off fifth bit
    cmp dl,'A'-48
    jb .LastDigit           ;character is less than 'A'
    cmp dl,'F'-48
    ja .LastDigit           ;character is greater than 'F' (15)
    sub dl,7                ;(65-48)+10  "A"=65->10  "F"=86->15
.AddPlaceValue:
    shl eax,4               ;multiply existing number by radix (16)
    inc esi                 ;move forwards one character
    or al,dl                ;add new digit
    jmp short .NextDigit
.LastDigit:

;(eax=number) (bl=letter)
    and bl,~32                  ;make uppercase
    cmp bl,'A'
    jne .SkipAdr
    mov [Sound.IoPort],eax      ;set DSP command reg and FM port
    add eax,byte SbAdr.Fm
    mov [Sound.FmPort],eax
.SkipAdr:
    cmp bl,'D'
    jne .SkipDma
    mov [Sound.Ldma],al         ;DMA Channels 0-3
.SkipDma:
    cmp bl,'H'
    jne .SkipHdma
    mov [Sound.Hdma],al         ;High DMA 4-7
.SkipHdma:
    cmp bl,'I'
    jne .SkipInt
    cmp al,16                   ;compensate for hexadecimal conversion
    jb .IrqOk
    sub al,6                    ;16->10, 21->15
.IrqOk:
    cmp al,2                    ;hack exception for cascaded IRQ 2
    jne .NotIrq2
    mov al,9                    ;force 2 to 9
.NotIrq2:
    mov [Sound.Irq],al
    ; get interrupt number based on IRQ
    cmp al,8
    jb .LowIrq
    and al,7
    add al,[Program.PicBaseSlave]
    jmp short .SetInt
.LowIrq:
    add al,[Program.PicBaseMaster]
.SetInt:
    mov [Sound.DaInt],al        ;keep interrupt number
.SkipInt:
    cmp bl,'P'
    jne .SkipMpu
    mov [Sound.MpuPort],eax
.SkipMpu:
;    cmp bl,'M'
;    jne .SkipMixer
;    mov [Sound.MixerPort],eax
;.SkipMixer:
;    cmp bl,'T'
;    jne .SkipType
;    mov [Sound..],eax
;.SkipType:
    jmp .NextElement

.End:
    clc
    ret


;
; Enables different waveforms, disables rhythm mode to make use of all 9
; channels.
;
; () ()
InitFmSound:
    test byte [SoundFlags],SoundFlags.Fm
    jz near .End
    and byte [SoundFlags],~SoundFlags.Fm
    mov esi,Text.InitializingFm
    call StatusMessage

    ; detect existance
    ; attempt to clear timer flags, then check if they are clear

    mov eax,0004h               ;stop timers
    call WriteFmReg
    mov eax,8004h               ;clear timer flags
    call WriteFmReg
    ;mov edx,[Sound.FmPort]
    dec edx
    in al,dx
    test al,0E0h                ;all timer flags should be clear
    jnz near .Fail

    ; attempt to start timer count, then check if they are set
    mov eax,0FF02h              ;set minimum timer 1 delay
    call WriteFmReg
    mov eax,0104h               ;start timer 1
    call WriteFmReg
    mov ecx,8                   ;delay at least 80 microseconds
    call TimedWait              ;delays 15-30us each count
    ;mov edx,[Sound.FmPort]
    dec edx
    in al,dx
    and al,0E0h                 ;mask only timer flags
    cmp al,0C0h                 ;timer 1 flag should be set
    jne .Fail

    mov eax,0004h               ;stop timers once again so they're not
    call WriteFmReg             ;just endlessly counting on
    mov eax,8004h               ;clear timer flags
    call WriteFmReg

    ; ack possible IRQ from timer overflow (unnecessary on most sound cards)
    mov al,[Sound.Irq]          ;needed for TeleSound 3D, or else audio
    or al,60h                   ;loops inifinitely after returning to Windows
    btr eax,3                   ;if IRQ 8-15
    jnc .LowIrq
    out Pic.Mode2,al
    mov al,62h                  ;specific EOI for chained IRQ 2
.LowIrq:
    out Pic.Mode1,al

.IgnoreDetect:                  ;call here to override
    ;mov ebx,4                  ;zero all registers (may be unnecessary)
    xor ebx,ebx
.NextRegZero:
    mov eax,ebx
    call WriteFmReg             ;destroys eax,ecx,edx
    inc bl                      ;next voice
    jnz .NextRegZero

    mov eax,2001h               ;enable different waveforms
    call WriteFmReg
    mov eax,00BDh               ;disable rhythm mode
    call WriteFmReg

; <The code below isn't really necessary since the patches will be set when
;  the notes are turned on anyway>
%if 0
    ; set default patch for all voices
    mov edi,9-1                 ;start with last voice
    mov esi,SbFm.PatchDefault
.NextVoicePatch:
    mov eax,edi
    call SetFmPatch
    dec edi
    jns .NextVoicePatch

    ; turn off all vaoices
    mov dword [SbFm.VoicesOff],-1
    mov ebx,00B0h               ;turn off voices
.NextVoiceOff:
    mov eax,ebx
    call WriteFmReg             ;destroys eax,ecx,edx
    inc ebx                     ;next voice
    cmp bl,0B8h
    jbe .NextVoiceOff
%endif

    cld
    mov edi,SbFm.UsedPatches
    mov eax,-1
    mov ecx,9
    rep stosd

    ; code should actually check for existance of FM instead of just
    ; assuming that it exists, since some cards actually lack FM synthesis.
    or byte [SoundFlags],SoundFlags.Fm
.End:
    ret

.Fail:
    mov esi,Text.FmNotReset
    jmp StatusMessage
    ;ret

;
.Release:
    btr dword [SoundFlags],SoundFlags.FmBit
    jnc .End
    mov esi,Text.ReleasingFm
    call StatusMessage

    mov ebx,00B0h               ;turn off voices
.NextVoiceOff:
    mov eax,ebx
    ;mov ah,[SbFm.UsedFnumbers+ebx-0B0h] ;get previous fnumber/octave of voice, to prevent clicking sound
    ;and ah,~20h                 ;clear bit to set off
    call WriteFmReg             ;destroys eax,ecx,edx
    inc ebx                     ;next voice
    cmp bl,0B8h
    jbe .NextVoiceOff
    ret


;
;
; Quote from "Programming the AdLib/Sound Blaster" by Jeffrey S. Lee (1992)
;
; After writing to the register port, you must wait twelve cycles before
; sending the data; after writing the data, eighty-four cycles must elapse
; before any other sound card operation may be performed.
;
; The AdLib manual gives the wait times in microseconds: three point three
; (3.3) microseconds for the address, and twenty-three (23) microseconds
; for the data.
;
; The most accurate method of producing the delay is to read the register
; port six times after writing to the register port, and read the register
; port thirty-five times after writing to the data port.
;
; (al=sb reg, ah=value)
; (edx=fm port+1; ebx,esi,edi)
;
WriteFmReg:
    mov edx,[Sound.FmPort]
    out dx,al
    mov ecx,2
    shl eax,8       ;preserve register value in upper dword
    call TimedWait  ;delays 15-30us
    shr eax,16      ;restore register value from upper dword
    inc edx
    out dx,al
    mov ecx,3       ;delays 30-45us
    ;jmp TimedWait

;
; Quote from "CPU Independant Timer"
;
; "the perfect anwser.  It seems that there is a timer connected to port 61h
;  at bit 4.  This bit toggles every 15.085 u sec (micro sec)."
;
; (ecx=number of 15.085 microsecond loops)
; (; !ax,ecx)
;
TimedWait:
    in al,61h
    and al,10h          ;mask bit 4
    mov ah,al
.CheckAgain:
    in al,61h
    and al,10h
    cmp al,ah
    jz .CheckAgain      ;no change so loop back
    mov ah,al
    dec ecx
    jg .CheckAgain
    ret


%if 0
;
; Rather than just abruptly turn off all the notes, it first sets the release
; rates so that they fade out.
SilenceFm:
    test byte [SoundFlags],SoundFlags.Fm
    jz .End

    mov ebx,0080h
.Next:
    mov eax,ebx
    mov ah,054h                 ;sustain/release
    call WriteFmReg             ;destroys eax,ecx,edx
    mov ah,075h                 ;attack/decay
    call WriteFmReg             ;destroys eax,ecx,edx
    inc ebx                     ;next voice
    cmp bl,060h+31
    jbe .Next
.End:
    ret
%endif


;
; (esi=patch info, al=voice)
; (; esi, edi)
; Voices are zero based.
SetFmPatch:
    movzx eax,al
    xor ebx,ebx                 ;start with first attribute
    push dword [SbFm.VoiceTable+eax] ;save voice operator offset
    jmp short .FirstReg
.NextReg:
    add al,[esp]                ;add voice operator offset
    mov ah,[esi+ebx]            ;get voice attribute value
    inc ebx                     ;next voice attribute
    call WriteFmReg
.FirstReg:
    mov al,[SbFm.PatchRegs+ebx]
    test al,al
    jnz .NextReg
.End:
    pop eax                     ;release saved operator offset
    ret


;
; Resets/detects sound card, allocates sound buffer, and sets IRQ interrupt
; handler. If the sound card does not exist, is at the wrong address, will
; not reset (because some other program already has control of it), or the
; sound buffer could not be allocated, the routine will return failure.
;
; Does absolutely nothing to initialize DMA channels, nor does it send any
; commands the sound card for transfer. That is all taken care of by
; InitDmaTransfer.
;
; write 1 to reset
; wait 3us
; write 0 to reset
; check for 0AAh from read port
; get sound card DSP version
; allocate's wave buffer
; clear buffer
;
; () ()
InitDaSound:
    test byte [SoundFlags],SoundFlags.Da
    jz near .End
    and byte [SoundFlags],~(SoundFlags.Da|SoundFlags.HighSpeed|SoundFlags.NeedHighSpeed)
    mov esi,Text.InitializingWave
    call StatusMessage

    ; detect sound card's existance and willingness to reset
    mov edx,[Sound.IoPort]
    mov al,1
    add dl,SbAdr.Reset
    out dx,al                   ;write 1 to reset port
    mov ecx,2                   ;delays 15-30us (need 3us)
    call TimedWait
    xor al,al
    out dx,al

    add dl,SbAdr.ReadStatus-SbAdr.Reset
    mov ecx,67                  ;although really only 100 microseconds are
.Wait:                          ;needed, allow for timeout of 1 millisecond
    in al,dx                    ;67 = .001 / .000015085
    test al,al
    jns .NotReady
    add dl,SbAdr.Read-SbAdr.ReadStatus
    in al,dx
    sub dl,SbAdr.Read-SbAdr.ReadStatus
    cmp al,0AAh
    je .Success

.NotReady:
    in al,61h                   ; Read timer connected to port 61h at bit 4
    and al,10h                  ; This bit toggles every 15.085 (micro sec)
    cmp al,ah
    je .Wait
    mov ah,al
    dec ecx
    jg .Wait
.Fail:
    mov esi,Text.SbNotReset
    jmp StatusMessage
.End:
    ret

.Success:
    ; get sound card's DSP version
    mov al,SbCmd.GetVersion
    call WriteSbDsp
    call ReadSbDsp
    cmp al,2                    ;need at least SB2 for auto init
    jb .Fail                    ;and 2.1 for high speed

    cmp al,4                    ;need 4.xx for 16bit
    mov cl,[SoundFlags]
    jae .16bit
    or cl,SoundFlags.NeedHighSpeed
    test cl,SoundFlags.Supports16bit
    jnz .SetFlags
    or cl,SoundFlags.Force8bit
    jmp short .SetFlags
.16bit:
    or cl,SoundFlags.Supports16bit
.SetFlags:
    mov [SoundFlags],cl
    mov bh,al

    call ReadSbDsp
    mov bl,al
    mov [Sound.CardVersion],bx

    ; allocate sound buffer
    mov eax,0100h               ;allocate low memory
    mov ebx,(Sound.BufferSize*2)/16 ;paragraphs to grab
    int 31h
    jc .Fail
    shl eax,4
    mov [Sound.BufferSel],dx
    mov [Sound.BufferPtr],eax

    ; ensure buffer doesn't cross 64k boundary
    or eax,0FFFF0000h           ;set all bits above 64k (DMA page)
    add eax,Sound.BufferSize-1
    jnc .NoWrap
    add dword [Sound.BufferPtr],Sound.BufferSize
.NoWrap:

    ; get old interrupt handler
    mov ax,0204h                ;get protected-mode interrupt
    mov bl,[Sound.DaInt]        ;wave audio interrupt
    int 31h
    mov [.OldHandler],edx       ;save offset
    mov [.OldHandler+4],cx      ;save selector

    ; set new interrupt handler
    mov edx,SoundHandler
    ;mov bl,[Sound.DaInt]       ;wave audio interrupt
    mov ax,0205h                ;set protected mode interrupt
    mov cx,cs                   ;pass our code selector
    int 31h

    ; enable interrupt by setting IRQ bit in PIC
    mov cl,[Sound.Irq]
    mov bl,254
    mov edx,Pic.Mode1           ;default to PIC1, for IRQs 0-7
    cli
    cmp cl,8
    jb .EnableLowIrq
    mov al,60h|2                ;specific EOI, in case some other program left one
    out dx,al
    in al,Pic.Mask1             ;dx=Pic.Mask1
    and al,11111011b            ;unmask IRQ 2 also because of cascaded IRQs
    out Pic.Mask1,al            ;enable slave 8259 for high IRQs
    mov dl,Pic.Mode2            ;use PIC2 instead, for IRQs 8-15
.EnableLowIrq:
    mov al,cl
    and al,7
    or al,60h                   ;specific EOI, in case some other program left one
    out dx,al                   ;a few times after playing SpcTool then starting my own player, a spurious initial interrupt would occur
    inc dx                      ;mask register = mode reg + 1
    rol bl,cl                   ;enable (unmask) channel
    in al,dx
    and al,bl
    out dx,al
    mov [Sound.PicMaskPort],edx
    sti

    mov al,0D1h                 ;turn on speaker
    call WriteSbDsp

    or byte [SoundFlags],SoundFlags.Da
    ret

;
.Release:
    btr dword [SoundFlags],SoundFlags.DaBit
    jnc .IgnoreCleanUp
    mov esi,Text.ReleasingWave
    call StatusMessage

    mov al,0D3h                 ;turn off speaker to silence any low hiss
    call WriteSbDsp
    mov edx,[Sound.IoPort]      ;reset sound card one last time just in case
    mov al,1                    ;something was left unresolved
    add dl,SbAdr.Reset
    out dx,al

    ; disable interrupt
    mov edx,[Sound.PicMaskPort]
    mov cl,[Sound.Irq]
    cmp cl,9
    je .KeepIrq                 ;leave chained IRQ2 enabled or SpcTool 0.6
    mov bl,1                    ;won't know how to handle it (total silence)
    rol bl,cl
    in al,dx
    or al,bl
    out dx,al
.KeepIrq:

    ; restore old interrupt handler
    mov bl,[Sound.DaInt]        ;wave audio interrupt
    mov ax,0205h                ;set protected mode interrupt
    mov cx,[.OldHandler+4]      ;get selector
    mov edx,[.OldHandler]       ;get offset
    int 31h

    ; free buffer
    mov ax,0101h                ;deallocate low memory
    mov dx,[Sound.BufferSel]
    int 31h

.IgnoreCleanUp:
    ret


section .bss
alignb 4
.OldHandler:    resd 1
                resw 1
section .text


;
; Near pointless function whose sole function is to simply acknowledge the
; unused sound card generated interrupts and return. Not much, but it is very
; important for some computers. On mine, the sound simply stops playing,
; while some actually triple-fault! Yeah, the entire stupid computer reboots
; just because the sound card was not acknowledged. The dumbest thing is that
; it happens whether the IRQ channel is enabled or disabled.
SoundHandler:
    push eax
    push edx
    push ds

    ; increment character on screen for debugging
    mov ds,[cs:Program.DataSelector]
    ;inc dword [0B8000h+160]

    ; my PC freezes if the MPU is never acknowledged
    test byte [SoundFlags],SoundFlags.Mpu
    jz .NoMpu
    mov dx,[Sound.MpuPort]
    in al,dx
.NoMpu:

    ; Acknowledge the SB's interrupt so it doesn't send it again
    ; Doing this simple step below is gravely important. Without it, the sound
    ; card may continue sending endless more interrupts to the point that the
    ; stack overflows and the pc triple faults - most annoying!
%if 0
    mov edx,[Sound.IoPort]
    add dl,SbAdr.Ack8bit
    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bit
    ;add dl,SbAdr.Ack16bit-SbAdr.Ack8bit
    inc dl
 .8bit:
    in al,dx
%else
    mov edx,[Sound.IoPort]
    add dl,SbAdr.Ack8bit
    in al,dx
    inc edx
    in al,dx
%endif

    ; allow the PIC to signal more interrupts
    cmp byte [Sound.Irq],8
    mov al,20h
    jb .LowIrq
    out Pic.Mode2,al
.LowIrq:
    out Pic.Mode1,al
    pop ds
    pop edx
    pop eax
    iret


;
; Reads single byte from sound card. Many program's sound routines just have
; a tight loop here, so that if anything goes wrong, you're stuck in one of
; those great infinite loops. Personally, I can stand the thought of such
; code, so I provide a timeout by using the 15us timer. If timeout occurs,
; the routine ignores the fact that the card isn't ready and just reads it
; anyway.
;
; ()
; (al=data, ecx=success if nonzero; !edx)
ReadSbDsp:
    mov edx,[Sound.IoPort]
    add dl,SbAdr.ReadStatus
    mov ecx,67                  ;set timeout to 1 millisecond just in case
.Wait:                          ;67 = .001 / .000015085
    in al,dx
    test al,al
    js .Ready
    in al,61h                   ; Read timer connected to port 61h at bit 4
    and al,10h                  ; This bit toggles every 15.085 (micro sec)
    cmp al,ah
    je .Wait
    mov ah,al
    dec ecx
    jg .Wait

.Ready:
    add dl,SbAdr.Read-SbAdr.ReadStatus
    in al,dx
    ret


;
; Writes single byte to sound card. If timeout occurs, the routine ignores
; the fact that the card isn't ready and just writes it anyway.
;
; (al=data)
; (ecx=success if nonzero; !ecx,edx)
WriteSbDsp:
    push eax
    mov edx,[Sound.IoPort]
    add dl,SbAdr.WriteStatus
    mov ecx,67                  ;set timeout to 1 millisecond just in case
.Wait:                          ;67 = .001 / .000015085
    in al,dx
    test al,al
    jns .Ready
    in al,61h                   ; Read timer connected to port 61h at bit 4
    and al,10h                  ; This bit toggles every 15.085 (micro sec)
    cmp al,ah
    je .Wait
    mov ah,al
    dec ecx
    jg .Wait

.Ready:
    pop eax
    ;add dl,SbAdr.Write-SbAdr.WriteStatus  ;same port
    out dx,al
    ret


;
; Separate from InitDaSound so that it can be called multiple times. That way
; the sample rate, bits, and stereo options can be changed within the program
; without needing to exit and restart it. Supports 8bit, high speed 8bit, and
; 16bit, but all of them in mono only (didn't care enough for stereo).
;
; Note while the DMA controller is programmed with the true buffer size, the
; sound card is given a ridiculously large value so that it signals the
; minimum frequency of pointless interrupts, since none of the sound
; processing even occurs in the handler. The handler's only purpose is to
; simply acknowledge the interrupts and return.
;
; Oddly, some sound cards always generate at least one interrupt per buffer
; transfer, regardless of the transfer size given to it (Crystal Audio 1999).
; I think what is happening is that the legacy emulator is intercepting all
; sound related port writes (even to the DMA) and choosing to do whatever it
; wants. No big deal though...
;
InitDmaTransfer:
    test byte [SoundFlags],SoundFlags.Da
    jz near .IgnoreCall
    mov esi,Text.StartingSound
    call StatusMessage

    ; clear the sound buffer and reset its position offset
    mov dword [Sound.BufferPos],0
    call ClearSbBuffer.NoCheck

    ; The reason the convolution immediately below is that some sound cards
    ; setups can actually use the 8bit DMA channels for 16bit output. Another
    ; reason is that they sometimes 'forget' to set the H parameter in the
    ; BLASTER variable.

    mov ebx,[Sound.Hdma]
    cmp bl,4
    je .UseLdma
    test byte [SoundFlags],SoundFlags.Force8bit
    jz .UseHdma
.UseLdma:
    mov bl,[Sound.Ldma]
.UseHdma:
    mov [Sound.DmaChannel],bl
    cmp bl,4
    jae .SetupHdma

; (ebx=dma channel)
.SetupDma:
    ; set up transfer on DMA side
    mov al,bl
    or al,4
    out 0Ah,al                  ;mask DMA channel for programming
    ;mov al,bl                  ;just done above
    xor al,58h | 4
    out 0Bh,al                  ;set transfer mode (single/autoinit/write)
    ;xor al,al                  ;unnecessary since any value will work
    out 0Ch,al                  ;reset flip-flop

    mov eax,[Sound.BufferPtr]
    lea edx,[ebx*2]             ;port for address and length
    out dx,al                   ;set address low byte
    mov al,ah
    out dx,al                   ;set address high byte
    inc dl
    mov al,(Sound.BufferSize-1) & 255
    out dx,al                   ;set buffer size low byte
    mov al,(Sound.BufferSize-1) >> 8
    out dx,al                   ;set buffer size high byte
    mov dx,[.DmaPageTbl+ebx*2]  ;get page register for channel
    shr eax,16                  ;get top four bits of buffer ptr for page
    out dx,al                   ;set page
    mov al,bl
    out 0Ah,al                  ;reenable channel for transfer

    jmp short .SetupSb

;(ebx=dma channel)
.SetupHdma:
    ; set up transfer on DMA side
    mov al,bl
    ;or al,4                    ;bit is already set anyway
    out 0D4h,al                 ;mask DMA channel for programming
    ;mov al,bl                  ;just done above
    xor al,58h | 4
    out 0D6h,al                 ;set transfer mode (single/autoinit/write)
    ;xor al,al                  ;unnecessary since any value will work
    out 0D8h,al                 ;reset flip-flop

    mov eax,[Sound.BufferPtr]
    lea edx,[ebx*4+0C0h-16]     ;port for address and length
    shr eax,1                   ;both page and offset / 2
    out dx,al                   ;set address low byte
    mov al,ah
    out dx,al                   ;set address high byte
    add dl,2
    mov al,(Sound.BufferSize/2-1) & 255
    out dx,al                   ;set buffer size low byte
    mov al,(Sound.BufferSize/2-1) >> 8
    out dx,al                   ;set buffer size high byte
    mov dx,[.DmaPageTbl+ebx*2]  ;get page register for channel
    shr eax,15                  ;get top four bits of buffer ptr for page
    out dx,al                   ;set page
    mov al,bl
    and al,3
    out 0D4h,al                 ;reenable channel for transfer

.SetupSb:
    ; determine number of samples to generate per timer interrupt. A common
    ; rate like 22050hz would require 735 per interrupt (30 ints per second).
    mov eax,[Sound.SampleRate]
    xor edx,edx
    mov ebx,30                  ;PIT interrupts per second
    div ebx                     ;samples per sec = sample rate / frames
    mov [Sound.IntSamples],eax  ;constant samples to generate per interrupt
    mov [Sound.MixSamples],eax  ;variable samples to mix and send to sound card (depending on sync)

    ; tell the sound card the sampling rate
    cmp dword [Sound.CardVersion],400h  ;>= 4.x supports direct sample rate
    jb .SetTimeConstant         ;instead of dumb time constants
    mov al,41h
    call WriteSbDsp
    mov eax,[Sound.SampleRate]
    ;mov [Sound.MixRate],eax
    xchg al,ah                  ;high byte then low... why?
    call WriteSbDsp
    mov al,ah
    call WriteSbDsp
    jmp short .SampleRateSet
.SetTimeConstant:
    ; the SB documentation gives the formula: 65536-(256000000/SampleRate)
    ; but this is one slightly more accurate: 256-(1000000/SampleRate) using
    ; the first formula and playing this program alongside the real SNES
    ; sounds terrible (the off pitch hurts my ears :)
    mov al,040h                 ;set time constant
    call WriteSbDsp             ;TimeConstant = 65536-(256000000/SampleRate) or 256-(1000000/SampleRate)
    xor edx,edx                 ;11025hz=A5 22050=D2 32000=E0 44100=E9
    ;mov eax,256000000
    mov eax,1000000
    mov ebx,eax
    div dword [Sound.SampleRate]
    neg al                      ;same as 256-x
    ;neg eax                    ;same as 65536-x
    ;mov al,ah
    call WriteSbDsp
   %if 0                        ;doesn't seem to make any difference, the
    neg al                      ;pitch is still of at certain sampling rates
    imul dword [Sound.SampleRate]
    mov ebx,1000000
    imul dword [Sound.SampleRate]
    div ebx
    mov [Sound.MixRate],eax
   %endif

.SampleRateSet:

    ; set up transfer on SB side
    test byte [SoundFlags],SoundFlags.Force8bit
    jz .16bit

.8bit:
    mov al,048h                 ;set transfer size
    call WriteSbDsp
    ;mov eax,[Sound.IntSamples]
    mov al,Sound.TransferSize & 255
    call WriteSbDsp             ;low byte first
    ;mov al,ah
    mov al,Sound.TransferSize >> 8
    call WriteSbDsp             ;high byte next

    ; if DSP version < 4 and sample rate > 23000, enter high speed mode
    ; if you try play 23001-44100hz without entering high speed, the sample
    ; rate is clipped to its maximum 23000hz by the card.
    cmp dword [Sound.SampleRate],Sound.HighSampleRate
    jbe .NoHighSpeed
    test byte [SoundFlags],SoundFlags.NeedHighSpeed
    jz .NoHighSpeed             ;>= 4.x need not high speed mode
    or byte [SoundFlags],SoundFlags.HighSpeed
    mov al,90h                  ;begin high-speed 8bit PCM output
    jmp WriteSbDsp
    ;ret
.NoHighSpeed:
    mov al,01Ch                 ;begin 8bit PCM output
    jmp WriteSbDsp
    ;ret

.16bit:
    mov al,10110110b            ;begin 16bit PCM output, fifo on
    call WriteSbDsp
    mov al,10h                  ;mode=16bit signed mono
    call WriteSbDsp
    ;mov eax,[Sound.IntSamples]
    mov al,Sound.TransferSize & 255
    call WriteSbDsp             ;low byte first
    ;mov al,ah
    mov al,Sound.TransferSize >> 8
    jmp WriteSbDsp              ;high byte next
    ;ret

;
.Stop:
    test byte [SoundFlags],SoundFlags.Da
    jz .IgnoreCall
    mov esi,Text.StoppingSound
    call StatusMessage

    ; one dumb sound card (Crystal Audio Codec) does not stop playing when
    ; you TELL it to stop. No, you tell it to exit, pause, disable the DMA
    ; channel, or even reset it, and it still keeps playing! So, the least
    ; we can do is silence it.
    call ClearSbBuffer.NoCheck

    ; stop sound card transfer
    ; if in high speed mode
    ;   reset sound card
    ; elseif 8bit
    ;   pause 8bit dma
    ; else 16bit
    ;   pause 16bit dma
    ; endif
    btr dword [SoundFlags],SoundFlags.HighSpeedBt
    jnc .StopLowSpeed
    mov edx,[Sound.IoPort]      ;reset sound card to exit high speed mode
    mov al,1
    add dl,SbAdr.Reset
    out dx,al                   ;write 1 to reset port
    mov ecx,2                   ;delays 15-30us (need 3us)
    call TimedWait
    ;xor al,al
    ;out dx,al                   ;write 0 to reset
    ;mov ecx,7                   ;delay 100 microseconds
    ;call TimedWait              ;7 = .0001 / .000015085
    ;jmp short .Stopped
.StopLowSpeed:
    test byte [SoundFlags],SoundFlags.Force8bit
    mov al,SbCmd.PauseDma8bit
    jnz .Pause8bit
    mov al,SbCmd.PauseDma16bit
.Pause8bit:
    call WriteSbDsp
.Stopped:

;    test byte [SoundFlags],SoundFlags.Force8bit
;    mov al,SbCmd.ExitDma8bit
;    jnz .Exit8bit
;    mov al,SbCmd.ExitDma16bit
;.Exit8bit:
;    call WriteSbDsp

    ; disable DMA channel
    mov al,[Sound.DmaChannel]
    cmp al,4
    jb .DisableDma
    ;or al,4 ;unnecessary since bit 2 is already set
    out 0D4h,al                 ;mask DMA channel for programming
    ret
    ;jmp short .DmaDisabled
.DisableDma:
    or al,4
    out 0Ah,al                  ;disable DMA channel
.DmaDisabled:
.IgnoreCall:
    ret


section .data
align 2
.DmaPageTbl:    dw 87h,83h,81h,82h, 8Fh,8Bh,89h,8Ah
section .text


;
; Simply clears the sound card buffer (not mixing buffer). Used whenever
; silence should be heard, like when the sound card is first initialized
; and no song is playing yet, or when everything is paused.
;
ClearSbBuffer:
    test byte [SoundFlags],SoundFlags.Da
    jz .IgnoreCall
.NoCheck:
    cld
    mov edi,[Sound.BufferPtr]
    xor eax,eax                 ;0 is silence for 16bit
    test byte [SoundFlags],SoundFlags.Force8bit
    jz .Clear16bit
    mov eax,81818181h           ;128 (80h) is silence for 8bit
.Clear16bit:                    ;but for some odd reason, 81h reduces static
    mov ecx,Sound.BufferSize/4  ;hiss, the barely audible high pitch tone,
    rep stosd                   ;and faint crackling sounds
.IgnoreCall:
    ret

%if 0 ;saw wave
    mov edi,[Sound.BufferPtr]
    xor eax,eax
    mov ecx,Sound.BufferSize
.ClearNext:
    mov [edi],ah
    add eax,byte 111
    inc edi
    dec ecx
    jg .ClearNext
%endif


;
; Simply clears the 32bit mixing buffer (not sound card buffer). Used before
; mixing digital audio of voices and sample together
;
; () ()
ClearMixBuffer:
    xor eax,eax
    mov edi,MixBuffer
    ;mov ecx,[Sound.MixSamples]
    ;add ecx,byte 3
    ;shr ecx,2                   ;/4
    mov ecx,Sound.MixBufferSize/4
    cld
    rep stosd
    ret


;
; Ensures the output and DMA are synchronized, in case you switch away to
; Windows, or if the sound card does not have accurate timing (which I
; have found!). For most sound cards, the mixing can safely stay just ahead
; of the sound output (closer timing), but for at least one (Crystal Audio)
; any less than a 1/30 second causes annoying clicks. So, this routine
; ensures the mixing is always one interrupt's buffer length ahead of DMA.
; That 1/30 second lag isn't even noticeable though, except when both FM and
; digital audio are played simultaneously, making it audible that the FM
; attack comes before the wave.
;
; The mixing rate and the sound card's rate are not exactly the same since
; the sound generation is tied to the PIT for more accurate timing, rather
; than an imprecise output rate that varies from card to card.
;
; If either process drifts apart (mixing ahead of DMA, or DMA faster than
; mixing), the routine will routine return a suggested buffer size to bring
; them closer together. If somehow the two get too far apart from each other,
; it will resynchronize the buffer position to the current DMA address.
;
; The visual below is more for my sake than yours. Envisioning it all in my
; mind was a little complicated.
;
; We're fine, DMA is a full buffer behind us
;             |----------|
;   ------*---|----------*----------|
;      DMA            MIX @4268
;
; We've fallen behind somehow and DMA is catching up
; Mix larger buffers to make up for it
;  |---------------------|
;   -*--------*----------|----------*
;   DMA      MIX @4268         NEW MIX @12804
;
; The next interrupt, the distance is fine again
;                        |----------|
;   ----------|-*--------|----------*
;              DMA                 MIX @12804
;
; () (eax=DMA address, edx=suggest buffer size)
SyncToDma:
    ; read address from 8237 DMA controller (weird configuration)
    mov edx,[Sound.DmaChannel]
    xor eax,eax                 ;zero address
    cmp dl,4
    jae .Hdma
    mov ecx,eax                 ;zero left shift
    out 0Ch,al                  ;clear flip-flop (value is irrelavant)
    shl edx,1                   ;*2 address port
    jmp short .ReadAddress
.Hdma:
    lea edx,[0C0h+edx*4-16]     ;address port
    out 0D8h,al                 ;clear flip-flop (value is irrelavant)
    mov cl,1                    ;set left shift (address*2)
.ReadAddress:
    in al,dx                    ;get address low byte
    mov ah,al
    in al,dx                    ;get address high byte
    xchg al,ah                  ;correct high/low byte order

    ; get offset in buffer & calculate lag
    shl eax,cl                  ;*2 for 16bit channel, *1 for 8bit channel
    ;mov edx,[Sound.IntSamples]
    mov edx,[Sound.IntSamples]
    sub eax,[Sound.BufferPtr]   ;buffer offset = DMA address - base ptr
    mov ecx,edx
    shr ecx,2                   ;/4
    mov ebx,[Sound.BufferPos]
    add eax,ecx                 ;+25% hack for Crystal Audio
    and eax,Sound.BufferSize-1
    sub ebx,eax                 ;lag = buffer pos - trailing DMA pos
    ; this does not use cl from above because some sound cards utilize low
    ; DMA channels for 16bit data
    and ebx,Sound.BufferSize-1
    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bit
    shr ebx,1                   ;lag /2 for words
.8bit:

    ;mov [PrintDebugInfo.Value1],eax
    ;mov [PrintDebugInfo.Value2],ebx

    ; check deviation difference (BufferPos - DmaPos should be >=IntSamples)
    ; (eax=DMA buffer offset, ebx=lag, edx=int samples)
    ;mov [PrintDebugInfo.Value1],ebx
    sub ebx,edx                 ;lag < buffer size
    je .InSync                  ;just fine
    jb .Behind                  ;falling behind, DMA will eventually pass us
    cmp ebx,edx                 ;we're faster than DMA
    jb .Ahead

; somehow got way off (Alt+Tab probably to blame)
; (eax=DMA buffer offset, edx=buffer size)
.OffSync:
    ; resync buffer position ahead of what DMA is currently reading
    ;inc dword [PrintDebugInfo.Value4]
    mov ecx,edx                 ;Sound.IntSamples*2
    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bitResync
    shl ecx,1                   ;*2 if words, else bytes
.8bitResync:
    add ecx,eax                 ;Sound.MixSamples+Dma address
    and ecx,Sound.BufferSize-2  ;wrap within buffer to nearest word
    mov [Sound.BufferPos],ecx
    ret
.InSync:
    mov [Sound.PrevLagDif],ebx
    ret

.Behind:
    ;inc dword [PrintDebugInfo.Value2]
    xchg [Sound.PrevLagDif],ebx
    add ebx,[Sound.PrevLagDif]
    sar ebx,1

    sub edx,ebx
    cmp edx,Sound.MaxIntSamples
    jbe .BelowMax
    mov edx,Sound.MaxIntSamples
.BelowMax:
    ret

.Ahead:
    ;inc dword [PrintDebugInfo.Value3]
    xchg [Sound.PrevLagDif],ebx
    add ebx,[Sound.PrevLagDif]
    sar ebx,1

    sub edx,ebx
    cmp edx,Sound.MinIntSamples
    jae .AboveMin
    mov edx,Sound.MinIntSamples
.AboveMin:
    ret


;
; Copies mixing buffer to sound card buffer in low memory, converting the
; 32bit samples to either 8bit or 16bit, and clipping out of range samples.
;
; () ()
CopyMixBuffer:
    mov esi,MixBuffer
    mov ebx,[Sound.BufferPtr]
    mov edi,[Sound.BufferPos]
    mov ecx,[Sound.MixSamples]

    test byte [SoundFlags],SoundFlags.Force8bit
    jnz .8bitNext

.16bitNext:
    mov eax,[esi]
    and edi,Sound.BufferSize-2  ;wrap (to even word)
    cmp eax,32768
    jb .16bitValid
    cmp eax,-32768
    jae .16bitValid
    sar eax,31                  ;sign extend
    xor eax,32767
.16bitValid:
    mov [ebx+edi],ax
    add esi,byte 4
    add edi,byte 2
    dec ecx
    jg .16bitNext
    jmp short .End

.8bitNext:
    mov eax,[esi]
    and edi,Sound.BufferSize-1  ;wrap
    add eax,32768               ;toggle sign bit for conversion to 8bit
    add esi,byte 4
    test eax,0FFFF0000h
    jz .8bitValid
    sar eax,31                  ;sign extend
    not eax
.8bitValid:
    mov [ebx+edi],ah
    inc edi
    dec ecx
    jg .8bitNext

.End:
    and edi,Sound.BufferSize-1
    mov [Sound.BufferPos],edi
    ret


%if 0
;
; Transfer mixing buffer to low memory for sound card, converting wave buffer
; data to 8bit unsigned and clipping.
;
; (esi=source data, edi=destination, ecx=samples)
; (esi=new source ptr, edi=new dest ptr; ebx,edx)
TransferWave8bit:
.Next:
    mov eax,[esi]
    add eax,32768               ;toggle sign bit for conversion to 8bit
    add esi,byte 4
    test eax,0FFFF0000h
    jz .Valid
    sar eax,31                  ;sign extend
    not eax
.Valid:
    mov [edi],ah
    inc edi
    dec ecx
    jg .Next
    ret


;
; Transfer mixing buffer to low memory for sound card, converting wave buffer
; data to 16bit signed and clipping.
;
; (esi=source data, edi=destination, ecx=samples)
; (esi=new source ptr, edi=new dest ptr; ebx,edx)
TransferWave16bit:
.Next:
    mov eax,[esi]
    add esi,byte 4
    cmp eax,32768
    jb .Valid
    cmp eax,-32768
    jae .Valid
    sar eax,31                  ;sign extend
    xor eax,32767
.Valid:
    mov [edi],ax
    add edi,byte 2
    dec ecx
    jg .Next
    ret
%endif


;
; Simply generates a square wave of a given tone.
;
; () ()
GenerateSquareWave:
    mov eax,[.Peak]
    mov ecx,[Sound.MixSamples]
    mov edx,[.Acm]
    mov ebx,[.Mod]
    mov edi,MixBuffer

.NextSample:
    add [edi],eax
    add edx,ebx
    jnc .SameSign
    neg eax
.SameSign:
    add edi,byte 4
    dec ecx
    jg .NextSample

    mov [.Acm],edx
    mov [.Peak],eax
    ret

;32000/440=72 samples
;440/32000=0.01375

.Start:
    mov eax,[.Hertz]
    mov ebx,[.Vol]
    shl eax,17                  ;move to top of dword
    xor edx,edx
    div dword [Sound.MixRate]
    mov [.Peak],ebx             ;set initial peak phase to volume
    shl eax,15
    mov [.Acm],dword 0
    mov [.Mod],eax
    ret

;.Stop:
;    ret


section .data
align 4
.Hertz:         dd 440          ;A (max=32000)
.Vol:           dd 6800         ;good volume to constrast most samples
.Peak:          dd 6800
section .bss
alignb 4
.Mod:           resd 1
.Acm:           resd 1
section .text


;
; Detects MPU, resets it, and enters UART mode.
; Be sure reset the MPU before quitting the program, or else you may
; experience strange side effects like Window's sound effects looping
; infinitely??? Maybe because the MPU and wave audio share the same
; interrupt, and the sound driver becomes confused.
;
; ()
; ()
InitMpu:
    test byte [SoundFlags],SoundFlags.Mpu
    jz .End
    mov esi,Text.InitializingMpu
    call StatusMessage

    ; try to reset MPU
    call WaitMpu.Out
    jc .Fail
    mov al,MpuCmd.Reset
    out dx,al                   ;reset MPU
    dec edx
    in al,dx
    call WaitMpu.In             ;wait for success/fail response
    dec edx                     ;MPU data port
    in al,dx
    jc .Fail
    cmp al,MpuCmd.RespAck       ;successful reset?
    ;stc (anything - FFh should set carry)
    jne .Fail

    ; once reset, enter UART mode
    call WaitMpu.Out
    jc .Fail
    mov al,MpuCmd.Uart
    out dx,al                   ;enter UART mode
    call WaitMpu.In             ;wait for success/fail response
    dec edx                     ;MPU data port
    in al,dx
    jc .Fail
    cmp al,MpuCmd.RespAck       ;successful reset?
    ;stc (anything - FFh should set carry)
    jne .Fail

    ;or byte [SoundFlags],SoundFlags.Mpu
.End:
    ret

.Fail:
    and byte [SoundFlags],~SoundFlags.Mpu
    mov esi,Text.MpuNotReset
    jmp StatusMessage
    ;ret

;
; !call this before exiting the program!
.Release:
    btr dword [SoundFlags],SoundFlags.MpuBit
    jnc .End
    mov esi,Text.ReleasingMpu
    call StatusMessage
    call WaitMpu.Out
    mov al,MpuCmd.Reset
    out dx,al                   ;reset MPU
    ret


;
; ()
; (cf=error, edx=mpu cmd port; esi,edi,ecx)
;
; waits until either data is ready to be written or is available to read,
; depending on the bits set in the mask.
WaitMpu.In:
    mov eax,0FFFF0000h | (MpuCmd.MaskIn << 8)
    jmp short WaitMpu
WaitMpu.Out:
    ; set counter to 65535, and mask to either in or out bit
    mov eax,0FFFF0000h | (MpuCmd.MaskOut << 8)
WaitMpu:
    mov edx,[Sound.MpuPort]
    inc edx
.Next:
    in al,dx
    test al,ah
    jz .Done                    ;done, cf is clear
    sub eax,1<<16
    jnc .Next
    ;stc                        ;error
.Done: ;(cf=0)
    ret


;
; (al=value) (; ebx,esi,edi,al)
WriteMpu:
    mov ecx,eax
    call WaitMpu.Out
    jc .Err
    ;(edx=mpu port+1 from call)
    ;mov edx,[Sound.MpuPort]
    mov eax,ecx
    dec edx
    out dx,al                   ;send to data port
.Err:
    ret

;
; (esi=data string ptr, ecx=number bytes) (; ebx,edi)
WriteMpuString:
    cld
.Next:
    call WaitMpu.Out
    jc .Err
    ;(edx=mpu port+1 from call)
    ;mov edx,[Sound.MpuPort]
    lodsb
    dec edx
    out dx,al                   ;send to data port
    dec ecx
    jnz .Next
.Err:
    ret
