; Spc2Midi - Main routine
; SPC700 DSP to MIDI converter
; PeekinSot (c)2001
;
; By Dwayne Robinson (Peekin)
; Began 1999-11-21
; Updated 2001-11-28
; Compiled with NASM and WDOSX
; http://fdwr.tripod.com/snes.htm
; FDwR@hotmail.com
;
; Major Thanks to:
;   Ledi for transcribing the SF Sound Manual
;   _Demo_ for letting me see the SPCPLAY 2.0 source code
;   zsKnight for answering my annoying questions
;   Gau for his SPC document
;   CyberGoth for encouragement
;   AntiRes for an excellent SPC player
;   Authors of NASM and WDOSX for the great tools to create this program
;
;   And of course, the awesome SNES music writers (like Koji Kondo!)
;   who gave me a reason to write this
;
; Compiling:
;   It uses NASM 0.98 (Netwide Assembler)
;   And WDOSX 0.95 (Wuschel's DOS Extender)
;
;   nasm -f oldrdf spc2midi.asm -o spc2midi.exe
;   stubit -nowfse spc2midi.exe
;
; Files
;   spc2midi.asm    main program that calls everything else
;   file.asm        file related, such as loading spc's and zst's
;   settings.asm    reading/writing settings file
;
;   spcemu.asm      SPC700 CPU emulation core
;   dspemu.asm      DSP emulation, register reaction and recording
;   dspsim.asm      sound synthesis (DA/FM/MPU/Sine output)
;   dspmisc.asm     miscellaneous DSP related routines
;   disasm.asm      disassembly of opcodes
;
;   midi.asm        MIDI logging
;   sound.asm       generic sound routines (DA/FM/MPU)
;
;   gui.asm
;   gfx.asm         lines/borders/fonts
;   spcview.asm     display modes for notes playing
;
; DSP Write Buffering:
;   Unlike most SPC players which play directly as they emulate, this player
;   stores DSP writes to a large buffer, which is what it really plays from.
;   This allows the song to easily be skipped forward, backward, or to
;   any position you like without having to restart from the beginning. It
;   can also greatly increase speed on the second play (after the initial run
;   through), since the emulation has already finished and the program can
;   focus purely on playing from the buffer.
;
; FM Synthesis
;   Another speed advantage for slow machines is that it can play only a few
;   synthesized notes every 1/30th of a second rather than up to 8*32000 wave
;   samples per second. Admittedly these simplified sounds don't do justice
;   to the mood of the real music though :( None of those beautiful samples,
;   no smooth slides, cool panning... Yep, take all those away, and what do
;   you have, a MIDI! ;)
;
; Sample Identification
;   Unlike SpcTool, Spc2Midi does not assign attributes to a sample's
;   directory entry (a number 0-255), instead it assigns them to the sample.
;   Each valid sample is uniquely identified by its length, loop length, and
;   a 32bit checksum. So that means if you set the attributes for a sample in
;   one SPC of Zelda, those same attributes automatically apply to that same
;   sample in another SPC, regardless of where it may have been remapped to
;   in the directory table.
;
; Human Hearing:
;   In human hearing, a note that is twelve semitones above another (or one
;   whole octave), is actually twice as high in frequency. If a note played
;   at the rate of 4096 (32khz) sounded a middle C, 2048 would be one octave
;   below, and 8192 would be an octave above. This exponential curve must be
;   converted to a linear scale to obtain the MIDI note.
;
; Converting Play Rates
;   BRR sample play rates (0-16383) can't be converted directly to MIDI notes,
;   since each sample may be recorded at its own pitch. For example, a flute
;   instrument sound may be recorded at middle C, but a piano sound might be
;   recorded at G. I'm not sure why the music programmers choose these
;   seemingly arbitrary pitches rather than make them all have the same base
;   pitch. To compensate for this though, the play rates are multiplied by
;   their sample's apparent pitch. This is the pitch as heard by the human
;   ear (when played at actual recording speed of 32khz) rather than some
;   other frequency determinant like fourier analysis. This product (after
;   being shifted back down because of the multiplication) indexes into a
;   large table, to convert the exponential value into a linear note number.
;   The default apparent pitch for all samples is 440hz (A5).
;
;   PitchValue = (PlayRate * ApparentPitch) >> 12 [4096]
;   NoteNumber = HertzToNoteTable [PitchValue]
;
; Hertz to Pitch Table
;   MIDI notes (C0-G10 or 0-127) range in frequency from 8hz to 12543hz, with
;   any pitches outside that range constrained. So, it would make sense to
;   build a table from 0-12543, but because the lower frequencies are so
;   close (notes 4 and 5 can both be rounded to 10hz), I'd rather double the
;   table size for more precision.
;
;   PitchValue = (PlayRate * (ApparentPitch*2)) >> 13 [8192]
;
;=============================================================================

%define ProgVersion ".116"
SpcDebugMode        equ 0   ;1=disassembly 2=graphical code path 3=dsp writes as pixels
SpcDebugIoPrint     equ 0
ShowDebugInfo       equ 0
PlayInInterrupt     equ 1   ;set zero for debugging sound routines (crashing in any INT8 routine is a really BAD thing to do!)
SupportPitchSlide   equ 1

CyclesPerSecond     equ 2496000
CyclesPerTickDef    equ 39      ;default cycles per SPC700 timer tick
TicksPerSecond      equ 63990   ;64000 is the true speed, but isn't evenly divisible by 30
TicksPerInt         equ TicksPerSecond/30   ;ticks to advance for every timer interrupt (1.1931817MHz / 30)
MaxPlayLag          equ 6       ;maximum allowable time difference between main loop and interrupt play (in interrupt counts)
DspBufferSize       equ 262144*4;megabyte buffer to store DSP writes

Opr.None            equ 0
FilePathMaxLen      equ 256

StackReg:
.Di     equ 0
.Si     equ 4
.Bp     equ 8
.Sp     equ 12
.Bx     equ 16
.Dx     equ 20
.Cx     equ 24
.Ax     equ 28

Screen:
.Height     equ 200
.Width      equ 320
.Adr        equ 0A0000h
.Bytes      equ .Height * .Width
.Buffer     equ .Adr


;
section .bss
; global program variables
Program:
.DataSelector:      resw 1  ;necessary for sound interrupt handler
.PspSelector:       resw 1  ;selector to PSP
.ParamCount:        resd 1  ;number of parameters
.ParamTable:        resd 1  ;array of pointers to each one
.Env:               resd 1  ;command environment strings
.PicBaseMaster:     resb 1
.PicBaseSlave:      resb 1

; real mode register structure
Regs:
.edi                resd 1
.esi                resd 1
.ebp                resd 1
.res                resd 1  ;Reserved by system
.ebx                resd 1
.edx                resd 1
.ecx                resd 1
.eax                resd 1
.flags              resw 1
.es                 resw 1
.ds                 resw 1
.fs                 resw 1
.gs                 resw 1
.ip                 resw 1
.cs                 resw 1
.sp                 resw 1
.ss                 resw 1

SpcFileName:        resb FilePathMaxLen
SpsFileName:        resb FilePathMaxLen

alignb 4
Rams:               ;FileRams must exactly match Rams!
SpcRam:             resb 65536
DspRam:             resb 128
SpcEram:            resb 64
RamsSizeOf          equ $-SpcRam

FileRams:           ;FileRams must exactly match Rams!
FileSpcRam:         resb 65536
FileDspRam:         resb 128
FileSpcEram:        resb 64

DspRamCopy:         resb 128
FileSpcRegs:        resb 48 ;this does not need a copy

alignb 4
; Holds all the notes generated from KeyOn and KeyOff writes to the DSP.
; Once the buffer is full, no more of the song can be played and emulation
; stops, but the buffer size should be large enough to hold any complete song.
; The longest song (non-repeating) I've heard so far is the Mario RPG ending,
; which fits fine. Vortex has really long songs too. Let me know IF you find
; any nonrepetitive ones that don't fit, stopping short of the end.
DspBuffer:          resb DspBufferSize

; Information/attributes for all the sound effects used in a game. I would
; be very surprised if any game contained more than 512 instrument sounds.
SampleList:
.Total              equ 512
.Used:              resb .Total/8 ;bit flags to indicate which entries contain information
.ClearSize          equ $-SampleList
.Length:            resd .Total ;sample length (not bytes)
.LoopLen:           resd .Total ;length of loop portion
.Checksum:          resd .Total ;addition checksum of both sample and loop
.BrrIdx:            resd .Total ;matching BRR sample (if any)
.Instrument:        resd .Total ;MIDI instrument or patch (0-127)+(bank<<8)
.SoundFx:           resd .Total ;FM sound effect to use, if -1 then default for midi instrument is used
.Freq:              resd .Total ;closest frequency when sample played at 32khz
.Volume:            resb .Total ;volume relative to full MIDI volume
.Interpolation:     resb .Total ;simply for possible compatibility with SpcTool
.NameSize           equ 32
.NameShl            equ 5
.Name:              resb .Total*.NameSize
.SizeOf             equ $-SampleList

DefaultSample:
.Length             equ 0
.LoopLen            equ 0
.CheckSum           equ 0
.BrrIdx             equ -1
.Instrument         equ 46 ;46=orchestral strings
.SoundFx            equ 0
.Freq               equ 440
.Volume             equ 127
.Interpolation      equ 0

; Information about each sound effect currently in SPC memory, pointed
; to by the DSP table. Each sample may have other attributes associated with
; it from the SampleInfo list by an index. The reason why an index into a
; separate table is used rather than storing all the attributes with each
; sample, is because some sample numbers share the same BRR wave sample.
; Also, because different parts in the game may use the same sound effects
; but map them to different numbers, this separation allows those samples to
; always be correctly identified regardless of where they in the BRR table.
BrrList:
.Used:              resb 256/8  ;256 bit flags of samples have been used by music code
.Cached             resb 256/8  ;which samples have been ADPCM decompressed
.Valid              resb 256/8  ;which samples seem valid based on various rules (even if they are not)
.SampleIdx:         resd 256    ;ptrs into SampleList for more attributes about each sample
.SamplePtr:         resd 256    ;pointer to cached sample in the buffer
.LoopPtr:           resd 256
.Length:            resd 256    ;wave length in samples (not bytes), 0 if invalid sample
.LoopLen:           resd 256    ;length of loop portion in samples, 0 if no loop portion
.ClearSize          equ $-BrrList
.Offset:            resd 256    ;word offsets of sample:loop
.Checksum:          resd 256    ;addition checksum of both sample and loop, any trailing odd bytes are zero extended

SamplesBufferSize   equ 232992  ;64k * (8/9) * 4
SamplesBuffer:      resb SamplesBufferSize

HertzNoteTbl:       resb 12543*2+1  ;to convert frequency in hertz to MIDI note (0-127)

; These structures hold note information for the current time position in the
; song. That info includes the pitch, volume, key on status, and sample
; number of each voice, along with an complete copy of DSP RAM. The static
; structure is updated after each emulation run and is used mainly by the
; different display modes to reflect the notes playing to the screen. The
; real time structure is used by the sound playing interrupt routine. The
; reason why there are separate structures is so that each can be independant
; of each other. MPU output is not dependant on wave output. FM output is not
; dependant on the visual display. Also, it would be a mess if the sound
; routine messed up the display window right in the middle of redrawing.
alignb 4

struc SongPos
.DspRegs:           resb 128
.Pos:               resd 1      ;ptr to current command
.Time:              resd 1
.Active:            equ 4Ch     ;voices currently on (same as DSP register)
.Kon:               resb 1      ;voices turned on
.Kof:               resb 1      ;voices turned off
.Changed:           resb 1      ;1  voices turned either on or off
.Muted:             resb 1      ;1  voices currently muted by user
.Pitch:             resb 8
.Volume:            resd 8      ;combined volume of left & right channels
.Instrument:        resb 8
.SizeOf:            ;~164
endstruc

%macro NewSongPos 1
%1:
.DspRegs            equ $+SongPos.DspRegs
.Pos                equ $+SongPos.Pos
.Time               equ $+SongPos.Time    ;time in main loop (might fall behind real time)
.Kon                equ $+SongPos.Kon
.Kof                equ $+SongPos.Kof
.Active             equ $+SongPos.Active
.Changed            equ $+SongPos.Changed
.Muted              equ $+SongPos.Muted
.Pitch              equ $+SongPos.Pitch
.Volume             equ $+SongPos.Volume
.Instrument         equ $+SongPos.Instrument
.SizeOf             equ $+SongPos.SizeOf
                    resb SongPos.SizeOf
%endmacro

NewSongPos MainSongPos
NewSongPos FmSongPos
NewSongPos WaveSongPos
NewSongPos MpuSongPos
NewSongPos SineSongPos

;
section .data

align 4
PlayOptions:
.Timed              equ 1       ;play real speed, otherwise as fast as possible
.Emulate            equ 2       ;emulate SPC or stop emulation (when the buffer is full)
.FmVoices           equ 4       ;play using FM synthesis
.AsVoices           equ 8       ;play using digital audio samples (like any other SPC player)
.MpuVoices          equ 16      ;send MIDI commands to MPU
.SineVoices         equ 32      ;simulate with sine waves
.DlsVoices          equ 64      ;simulate with downloadable sounds
.Sample             equ 128     ;play the selected BRR sample
.MidiLog            equ 256     ;log to a MIDI file
.Interrupted        equ 512     ;flag that code was interrupted by timer
.Interruptedb       equ 9
.InInterrupt        equ 1024    ;flags we're currently in the interrupt to
.InInterruptb       equ 10      ;prevent possible recursion on slow machines
.TuningWave         equ 2048
.NoGui              equ 4096
.ShowInfo           equ 8192    ;display sample info
.WaveLog            equ 16384   ;record wave output to file
                    dd .Emulate ; | .FmVoices | .AsVoices | .MpuVoices | .SineVoices
PlayTimeStart:      dd 0        ;time before interrupt (start of time range)
PlayTimeEnd:                    ;time after interrupt (end of time range) synomymous with PlayTime
PlayTime:           dd 0        ;actual time (based on timer interrupt)
PlayTimePrev:       dd 0        ;previous play time
PlayStall:          dd 0        ;counter telling sound handler to wait because main loop is behind
PlayAdvance:        dd TicksPerInt ;number of ticks to advance each timer interrupt (may be 0 if paused)
PlaySpeed:          dd TicksPerInt ;number of ticks every 1/30 of a second

DspBufferEnd:       dd 0        ;pointer in buffer to last write
DspBufferTime:      dd 0        ;time of last event written to buffer
DspBufferKonTime:   dd 0        ;time of last key on written to buffer
DspBufferMinRes:    dd -1       ;minimum resolution between writes
DspFlags:           dd 0
.NewValidVoice      equ 1
.NewValidVoiceb     equ 0
DspFileHandle:      dd 0
CyclesTotal:        dd 0        ;total number cycles emulated so far
OpsTotal:           dd 0        ;total number of instructions executed
InterruptAccum:     dd 0        ;keeps track of when to chain to original int 8 handler

MutedVoices:        dd 255      ;default all voices active (1=active 0=muted)

SampleList.Entries:  dd 1       ;total number of entries in sample info list  >=1
SampleList.LastEntry:dd 0       ;total number of entries in sample info list  >=1
SoundFxList.Entries: dd 0       ;total sound effects  >=0

SelectedSample:     dd 0
DefaultGmPatch:     dd DefaultSample.Instrument

align 2
;HertzTbl:          incbin "gm2hzlut.dat"   ;130 words of nearest pitches
HertzNoteRle:       incbin "hz2gmlut.dat"   ;word rle encoded
GmFnumberTbl:       incbin "gm2fnlut.dat"
HertzFnumberTbl:    incbin "hz2fnlut.dat"
SineWaveTbl:        incbin "sinewave.dat"   ;1024 sample values

GuiPalette:         db 0,0,0,   20,15,23,30,24,34,41,38,46,52,48,55,63,63,63
                    db 36,26,42,40,30,60,20,20,40,20,20,60, 0,40,0,  0,60,0
                    db 40,40,0, 63,60,0, 40,0,0,  63,0,0
                    incbin "rainbow.pal",48
RainbowColorTable:  db 16,46,76,106,136,166,196,226
                    db 31,61,91,121,151,181,211,241


;
section .text
bits 32
global _WdosxStart

_WdosxStart:
;
Main:
    mov [Program.ParamCount],esi    ;save the number of parameters,
    mov [Program.ParamTable],edi    ;array of pointers to each one,
    mov [Program.Env],ebp           ;and command environment strings
    mov [Program.PspSelector],es
    mov [Program.DataSelector],ds
    mov ax,0400h                    ;get DPMI version
    int 31h
    mov [Program.PicBaseMaster],dh  ;interrupt chip base virtualization
    mov [Program.PicBaseSlave],dl   ;they are probably always 08h and 70h
    push ds
    pop es
    mov [Regs.sp],dword 0           ; tell DPMI host to provide stack
    ;mov [Regs.ss],word 0

    ; display title and set Window's title
    ;mov esi,Text.WinTitle
    ;call SetWindowsTitle
	mov edx,Text.Title
	call WriteString

    mov byte [SpcFileName],0    ;set initially no filename
    call CheckStartParams
    jnc .ParamsOk
.TerminateWithMsg:              ; (edx=message text ptr)
	call WriteString
    mov eax,4C01h
	int 21h
.ParamsOk:
    cmp byte [SpcFileName],0    ;check if filename present
    mov edx,Text.About
    je .TerminateWithMsg

    call LoadSpcState
    jc .TerminateWithMsg        ;end program and print error message

    mov esi,SpcFileName
    call GetFileName
    call SetWindowsTitle

    ; set screen row/col to 80x50
    mov ax,1112h                ;ah=set character font, al=8x8
    xor bx,bx                   ;table 0
    int 10h                     ;Video BIOS call
    mov ah,1                    ;set cursor type
    mov ecx,2000h               ;turn off cursor for faster disassembly
    int 10h                     ;Video BIOS call

   %if SpcDebugMode=2
	call ShowOpcode.PathInit
   %elif SpcDebugMode=3
    call ShowOpcode.SetVideoMode
   %endif

    ; load configuration file...
    ;get filename without extension and add .sps
    ;call LoadSpcSettings
    ;mov esi,Text.LoadingSettings
    ;call StatusMessage

    ; initialize core
    call DspEmu.Init            ;clear brr sample table, flag all samples as unused, expand 128 MIDI notes into 22k pitch table
    call SpcEmu.Restart         ;copy RAM, registers, set timers
    call GetAllSamplesInfo
    ;call DecompressAllSamples
    call MatchBrrSamples
    test dword [PlayOptions],PlayOptions.ShowInfo
    jz .NoInfo
    call DumpSoundBuffer.Info
    jmp .EndPremature
.NoInfo:

    ; initialize song seek structures
    mov edi,MainSongPos
    call SeekSongPos.Init       ;used by FM output, note display, and voice
    mov esi,MainSongPos         ;emulation routines. updated during fast
    mov edi,FmSongPos           ;forward/reverse.
    call SeekSongPos.Copy
    mov edi,WaveSongPos
    call SeekSongPos.Copy
    mov edi,MpuSongPos
    call SeekSongPos.Copy
    mov edi,SineSongPos
    call SeekSongPos.Copy

    ; initialize FM/digital audio
    call GetBlasterVars
    call InitDaSound            ;enable digital audio (for wave samples)
    call InitFmSound            ;enable waveform select, disable rhythm mode
    call InitMpu                ;enable MIDI port output
    call InitDmaTransfer

    ; Ensure at least one type of sound output is playing.
    ; For example, if FM could not be initialized for whatever reason, then
    ; use digital audio instead. Otoh, if audio will not reset, use FM.
    ; If any output mode is already set from command line, nothing changes;
    ; otherwise, the preference is digital audio, then FM.
    mov eax,PlayOptions.FmVoices
    test byte [SoundFlags],SoundFlags.Fm
    jz .NoFmInit
    test dword [PlayOptions],eax
    jnz .SoundOutputSet
.NoFmInit:
    test byte [SoundFlags],SoundFlags.Da
    jz .NoDaInit
    test dword [PlayOptions],PlayOptions.AsVoices|PlayOptions.SineVoices
    jnz .SoundOutputSet
    mov eax,PlayOptions.AsVoices
.NoDaInit:
    or dword [PlayOptions],eax
.SoundOutputSet:

    mov edx,Text.Ready
    call WriteString
    xor eax,eax
    int 16h
    cmp al,27
    je .End
    cmp al,'p'
    je .InitiallyPaused
    or byte [PlayOptions],PlayOptions.Timed
.InitiallyPaused:

    ;mov eax,168Fh               ;Win32 VM interrupt
    ;mov edx,0001h               ;enable application close
    ;int 2Fh

   %if SpcDebugMode=0
    test dword [PlayOptions],PlayOptions.NoGui
    jnz .NoInitGui
    ;call GuiInit
    mov eax,13h                 ;set graphics mode 13h
    int 10h
    mov esi,GuiPalette
    call SetPalette
    call SpcViewSetMode
.NoInitGui:
   %endif

    call SetupTimers            ;hook timer interrupt for sound
    call MainLoop
    call SetupTimers.Release

   %if SpcDebugMode=0
    test dword [PlayOptions],PlayOptions.NoGui
    jnz .NoExitGui
    ;call GuiDeinit
    mov eax,3                   ;set text mode
    int 10h
    ; set screen row/col to 80x50
    mov ax,1112h                ;ah=set character font, al=8x8
    xor bx,bx                   ;table 0
    int 10h                     ;Video BIOS
.NoExitGui:
   %endif

.End:
    call InitDmaTransfer.Stop
    call InitMpu.Release
    call InitFmSound.Release
    call InitDaSound.Release

.EndPremature:
   %if (SpcDebugMode=2) | (SpcDebugMode=3)
    mov eax,3               ;restore text mode
    int 10h
   %endif
    mov ah,1                   ;set cursor type
    mov ecx,0708h              ;turn on cursor
    int 10h

    ; save configuration file...

;
.Terminate:
    mov eax,4C00h
	int 21h

;
; do
;   if CyclesEmulated < CyclesToEmulate then
;     emulate for short time
;     update window if changed
;     if recording midi then record to midi file (not done)
;   endif
;   check keypresses and act on them
; loop
;
; note the sound is played in the interrupt
;
MainLoop:
    call ShowMutedVoices
    xor eax,eax

.NextEmuRun:
    test byte [PlayOptions],PlayOptions.Emulate
    jz .NoEmu
   %if SpcDebugMode=1
    mov eax,2000  ;for disassembly
   %else
    mov eax,83200 ;for normal playing (~1/30 second cycles)
   %endif
    call SpcEmu.Start
    btr dword [DspFlags],DspFlags.NewValidVoiceb
    jnc .NoNewVoices
    call MatchBrrSamples.New
.NoNewVoices:
    cmp dword [DspBufferEnd],DspBufferSize    ;stop emulation if buffer full
    jb .MoreEmu
    and byte [PlayOptions],~PlayOptions.Emulate
.NoEmu:
    ;hlt <- Win2k hates this multitasking, energy saving instruction!?@#
    ;Works great on Win95/98 reducing the CPU usage (on my PC anyway) from
    ;100% to only 12%.
    ;Does anyone else out there happen to think that every version of Windows
    ;becomes better in all the ways that don't matter, and worse in the
    ;aspects that matter most!? I'm beginning to hate "user-friendly".
.MoreEmu:

    ;test byte [PlayOptions],PlayOptions.MidiLog
    ;jz .NoMidiOutput
    ;call OutputMidiEvents
.NoMidiOutput:
    btr dword [PlayOptions],PlayOptions.Interruptedb
    jnc .NoUpdate
   %if ShowDebugInfo
    call PrintDebugInfo
   %endif

   %if SpcDebugMode=0
    mov edx,[PlayTime]

    test dword [PlayOptions],PlayOptions.NoGui
    jnz .NoGui

    ; advance position and update song structure
    cmp [MainSongPos.Time],edx
    je .NoChange
    mov edi,MainSongPos
    call SeekSongPos            ;seek time in edx
    ;mov edi,MainSongPos
   %if SupportPitchSlide
    call SeekSongPos.SetAllKeys ;all currently on voices
   %else
    call SeekSongPos.SetKeys    ;new keys only
   %endif

    ; draw next frame to song visualization
    call dword [SpcView.Handler]
    call dword [SpcView.ScreenCopier] ;this should later be moved to the GUI redraw
    jmp short .AfterGui

; (edx=play time)
.NoGui:
    mov eax,edx
    mov esi,Text.CurTime
    call .TimeToString
    mov eax,[DspBufferTime]
    mov esi,Text.CurTime+13
    call .TimeToString
    mov edx,Text.CurTime
    call WriteString
.AfterGui:

   %endif
.NoChange:

    ; play sounds da/fm/sine/mpu
   %if PlayInInterrupt=0
    call PlaySounds
   %endif
.NoUpdate:

    mov [PlayStall],dword 0

    ; check if user clicked on Window close button
    ;mov eax,168Fh               ;Win32 VM interrupt
    ;mov edx,0100h               ;query close
    ;int 2Fh
    ;jc .NoQueryClose            ;function not supported
    ;cmp al,1                    ;close application
    ;ja .NoQueryClose
    ;ret
.NoQueryClose:

    ; the key processing code below is presently an innefficient nightmare
	mov ah,1
	int 16h
    jz near .NextEmuRun

    push dword .NextEmuRun
    xor eax,eax
    int 16h

; just testing here with dynamically changing the playrate
%if 0
    cmp al,'z'
    jne .NotHighSampleRate
    call InitDmaTransfer.Stop
    mov [Sound.SampleRate],dword 22050
    jmp InitDmaTransfer
    ;ret
.NotHighSampleRate:
    cmp al,'x'
    jne .NotMidSampleRate
    call InitDmaTransfer.Stop
    mov [Sound.SampleRate],dword 32000
    jmp InitDmaTransfer
    ;ret
.NotMidSampleRate:
    cmp al,'c'
    jne .NotLowSampleRate
    call InitDmaTransfer.Stop
    mov [Sound.SampleRate],dword 43000
    jmp InitDmaTransfer
    ;ret
.NotLowSampleRate:
%endif

    cmp ax,'M'<<8
    mov ecx,TicksPerSecond
    je .ChangePos
    cmp ax,'K'<<8
    jne .NotChangePos
    neg ecx
.ChangePos:
    add ecx,[PlayTime]
    jns .PosOk
    xor ecx,ecx
.PosOk:
    mov [PlayTime],ecx
    ret
.NotChangePos:
    cmp al,']'
    mov ecx,TicksPerInt
    je .ChangePos
    neg ecx
    cmp al,'['
    je .ChangePos
.NotFineTune:
    cmp al,'9'
    ja .NotChannel
    je .TurnOnAll
    cmp al,'0'
    jb .NotChannel
    je .TurnOffAll
    movzx eax,al
    sub eax,byte '1'
    btc [MutedVoices],eax
    jnc .TurnedOn
    bts dword [SbFm.VoicesOff],eax
.TurnedOn:
    jmp ShowMutedVoices
    ;ret
.TurnOffAll:
    mov byte [MutedVoices],0
    mov byte [SbFm.VoicesOff],255
    jmp ShowMutedVoices
    ;ret
.TurnOnAll:
    mov byte [MutedVoices],255
    jmp ShowMutedVoices
    ;ret
.NotChannel:
    cmp ax,'H'<<8
    mov ebx,1
    je .ChangeSample
    cmp ax,'P'<<8
    jne .NotChangeSample
    neg ebx
.ChangeSample:
    add [SelectedSample],ebx
    ;and dword [SelectedSample],SampleList.Total-1
    and dword [SelectedSample],255
    mov esi,[SelectedSample]
    and dword [PlayOptions],~PlayOptions.Sample
    bt dword [BrrList.Valid],esi
    jnc .ShowInstName
    call SampleEmu.Start
    mov ecx,[SelectedSample]
    jmp short .ShowInstName
.NotChangeSample:
    cmp al,'.'
    mov ebx,1
    je .ChangeInstrument
    cmp al,','
    jne near .NotInstrument
    neg ebx
.ChangeInstrument:
    ;add [SampleEmu.PlayRate],ebx
    ;test dword [PlayOptions],PlayOptions.Sample
    ;jz near .NextEmuRun
    ;call SampleEmu.ChangeRate
    ;ret
    mov ecx,[SelectedSample]
    mov al,[SampleList.Instrument+ecx*4]
    add al,bl
    and eax,127
    mov [SampleList.Instrument+ecx*4],al
.ShowInstName:
    mov ecx,[SelectedSample]
    movzx eax,byte [SampleList.Instrument+ecx*4]
    mov al,[SampleList.Instrument+ecx*4]
    call GetMidiPatchName
    push ecx ;length of text
    push esi ;ptr to text
    push dword 0 ;color black
    push dword 16|(100<<16) ;height 8/width 100
    push dword 132 ;row 132/column 0
    call DrawRect.NoClip
    add esp,byte 12

    push dword 140 ;row 132/column 0
    call BlitString
    add esp,byte 12

    mov byte [NumToString.FillChar],' '
    mov eax,[SelectedSample]
    mov edi,NumToString.Buffer
    push dword NumToString.DefMaxLen ;length of text
    push edi
    call NumToString
    push dword 132 ;row 132/column 0
    call BlitString
    add esp,byte 12
    ret
.NotInstrument:
    cmp al,' '
    jne .NotPauseResume
    jmp PauseResumeSounds
    ;ret
.NotPauseResume:
    cmp al,8
    jne .NotReverse
    neg dword [PlaySpeed]
    neg dword [PlayAdvance]
    ret
.NotReverse:
    cmp ax,'G'<<8
    jne near .NotRestart
    push dword [PlayOptions]
    and byte [PlayOptions],~PlayOptions.Timed  ;do not want an interrupt to occur right here
    mov [PlayTime],dword 0      ;reset time
    mov [PlayTimeStart],dword 0 ;reset time
    mov edi,MainSongPos         ;initialize seek structures
    call SeekSongPos.Init
    mov esi,MainSongPos
    mov edi,FmSongPos
    call SeekSongPos.Copy
    mov edi,WaveSongPos
    call SeekSongPos.Copy
    mov edi,MpuSongPos
    call SeekSongPos.Copy
    mov edi,SineSongPos
    call SeekSongPos.Copy
    mov byte [SbFm.VoicesOff],255
    call SpcViewSetMode.ClearTable
    mov dword [PlaySpeed],TicksPerInt
    mov dword [PlayAdvance],TicksPerInt
    pop dword [PlayOptions]
    ret
.NotRestart:
    cmp ax,'O'<<8
    jne .NotEnd
    mov eax,[DspBufferTime]
    mov [PlayTime],eax
    ret
.NotEnd:
    cmp al,'+'
    jne .NotIncSpeed
    mov ecx,[PlaySpeed]
    add ecx,TicksPerSecond/(30*27)
    jz .ChangeSpeed
    jnc short .ChangeSpeed
.NotIncSpeed:
    cmp al,'-'
    jne .NotDecSpeed
    mov ecx,[PlaySpeed]
    sub ecx,TicksPerSecond/(30*27)
    jc .NotDecSpeed
.ChangeSpeed:
    mov [PlaySpeed],ecx
    mov [PlayAdvance],ecx
    ret
.NotDecSpeed:
    cmp al,'f'
    jne .NotFmToggle
    xor byte [PlayOptions],PlayOptions.FmVoices
    ret
.NotFmToggle:
    cmp al,'a'
    jne .NotAsToggle
    xor byte [PlayOptions],PlayOptions.AsVoices
    ret
.NotAsToggle:
    cmp al,'d'
    jne .NotDlsToggle
    xor byte [PlayOptions],PlayOptions.DlsVoices
    ret
.NotDlsToggle:
    cmp al,'s'
    jne .NotSineToggle
    xor byte [PlayOptions],PlayOptions.SineVoices
    ret
.NotSineToggle:
    cmp al,'g'
    jne .NotMpuToggle
    xor byte [PlayOptions],PlayOptions.MpuVoices
    ret
.NotMpuToggle:
    cmp al,27                   ;Escape pressed?
    jne near .NotEscape
    pop eax                     ;discard return address
.NotEscape:
    ret

.TimeToString:
    mov byte [NumToString.FillChar],'0'

    mov ebx,TicksPerSecond      ;~64000 ticks per second
    xor edx,edx
    div ebx
    lea edi,[esi+7]
    mov ecx,3
    push eax
    mov eax,edx
    shr eax,6                   ;/64
    call NumToString.UsingDLen

    mov ebx,60                  ;seconds
    pop eax
    xor edx,edx
    div ebx
    lea edi,[esi+4]
    mov ecx,2
    push eax
    mov eax,edx
    call NumToString.UsingDLen

    pop eax                     ;minutes
    mov edi,esi
    mov ecx,3
    call NumToString.UsingDLen

    ret

;
PlaySounds:
    test byte [SoundFlags],SoundFlags.Da
    jz .NoWave
    call SyncToDma
    mov [Sound.MixSamples],edx
    call ClearMixBuffer

    ;mov edx,[Sound.MixSamples]
    ;mov [PrintDebugInfo.Value1],edx

    test dword [PlayOptions],PlayOptions.AsVoices
    jz .NoAs
    call VoiceEmu
.NoAs:
    test dword [PlayOptions],PlayOptions.Sample
    jz .NoSample
    call SampleEmu
.NoSample:
    test dword [PlayOptions],PlayOptions.TuningWave
    jz .NoTuning
    call GenerateSquareWave
.NoTuning:
    test dword [PlayOptions],PlayOptions.SineVoices
    jz .NoSine
    call SineEmu
.NoSine:
;   test dword [PlayOptions],PlayOptions.DlsVoices
;   jz .NoDls
;   call DlsEmu
;.NoDls:

    ;call SyncToDma
    ;push edx                    ;save new buffer size
    call CopyMixBuffer
    ;pop dword [Sound.MixSamples]
.NoWave:

    test byte [SoundFlags],SoundFlags.Fm
    jz .NoFm
    ;call PlayFmSample
    test byte [PlayOptions],PlayOptions.FmVoices
    jz .NoFm
    call FmEmu
.NoFm:

    test byte [SoundFlags],SoundFlags.Mpu
    jz .NoMpu
    test byte [PlayOptions],PlayOptions.MpuVoices
    jz .NoMpu
    call MpuEmu
.NoMpu:

    mov eax,[PlayTimeEnd]       ;end of previous time range becomes start of
    mov [PlayTimeStart],eax     ;next one
    ret


;
PauseResumeSounds:
    xor byte [PlayOptions],PlayOptions.Timed
    test byte [PlayOptions],PlayOptions.Timed
    jz PauseSounds
    ret

PauseSounds:
    ;silence fm channels
    ;silence audio buffer

    ;and dword [PlayOptions],~PlayOptions.PlaySample
    call ClearSbBuffer

    ret


;
; Saves older timer interrupt handler and sets new one.
;
SetupTimers:
    ;save existing interrupt handler
    mov ax,0204h                ;get protected-mode interrupt
    mov bl,8                    ;timer interrupt
    int 31h
    mov [.OldHandler],edx       ;save offset
    mov [.OldHandler+4],cx      ;save selector

    ;increase rate to more precise rate of 30hz rather than only 18.2
    ;mov cx,39772               ;1193182/30=39772.7 (1.1931817MHz / 30hz)
    mov cx,39600                ;this value syncs better with the real SNES
    call .SetPitRate

    mov edx,TimerHandler
;(edx=handler address)
.SetNewHandler:
    mov bl,8                    ;timer interrupt
    mov ax,0205h                ;set protected mode interrupt
    mov cx,cs                   ;pass our code selector
    int 31h
    ret

.Release:
    ;restore old interrupt handler
    mov bl,8                    ;timer interrupt
    mov ax,0205h                ;set protected mode interrupt
    mov cx,[.OldHandler+4]      ;get selector
    mov edx,[.OldHandler]       ;get offset
    int 31h

    ;restore default rate of 18.2065hz
    xor cx,cx
    call .SetPitRate
    ret

.SetPitRate:
    cli                 ;don't want any interrupt to mess this up below
    mov al,00110100b    ;(00:11:010:0) set counter 0, lsb/msb, mode 2, binary
    out 43h,al
    in al,61h           ;i/o delay
    mov al,cl
    out 40h,al          ;write counter low byte
    in al,61h           ;i/o delay
    mov al,ch
    out 40h,al          ;write counter high byte
    sti
    ret

section .bss
alignb 4
.OldHandler:    resd 1
                resw 1
section .text

;
; This is called upon interrupt 8 from the pc timer (30fps). If play time in
; the main loop lags too far behind, it's given a chance to catch up.
;
; I've tried my best to time this exactly to the speed of the real games, and
; must say it's quite precise (at least on my computer). Started the Zelda
; dark world music playing both on the SNES and computer, came back a few
; minutes later (microwaved my Ramen lunch), and Spc2Midi was only an
; acceptable 1/30 second behind.
;
; The routine will exit if any one of several conditions are true:
;   Play is untimed (either paused or as fast as possible)
;   Emulation is behind real time
;   GUI loop has fallen behind and not acknowledged interrupts
;   File logging has not finished (so that previous wave data is not lost)
;
; //only if emulation is up to or ahead of realtime and main loop is not
; //lagging behind
; if (Timed
; && PlayTime - TickCount < MaxPlayLag
; && PlayTime + TicksPerInterrupt < TickTotal) then
;     PlayTime += TicksPerInterrupt
;     if (playing sounds) then
;         save all registers
;         read note buffer up to current time and play sounds
;         restore all registers
;     endif
; endif
; if InterruptAccum + ModRatio < 2^32 then
;     handle interrupt
; else
;     pass interrupt to BIOS and let it handle it
; endif
;
TimerHandler:
    push ds
    push es
    push eax
    mov ax,[cs:Program.DataSelector]
    mov ds,ax

    ;...(increment GUI time counter)

    ;inc dword [0B8000h]

    ; check for recursion or a main loop that is too slow
    bts dword [PlayOptions],PlayOptions.InInterruptb
    jc near .NoChain
    or dword [PlayOptions],PlayOptions.Interrupted
    test byte [PlayOptions],PlayOptions.Timed
    jz .NoPlay
    ; check if main loop is behind (data may need to be logged to file first)
    cmp dword [PlayStall],MaxPlayLag
    ja .NoPlay
    mov es,ax
    inc dword [PlayStall]

    ; advance tick time
    mov eax,[PlayTime]
    cmp [SpcEmu.TotalTime],eax  ;don't play faster than the emulation
    jb .NoTimerAdvance
    add eax,[PlayAdvance]       ;get next time frame
    jns .CheckTime
    jc .CheckTime
    xor eax,eax                 ;tick time went negative
    jmp short .TimeOk
.CheckTime:
    cmp [DspBufferTime],eax     ;can't play ahead of the buffer end
    jae .TimeOk
    mov eax,[DspBufferTime]
    inc eax
.TimeOk:
    mov [PlayTime],eax          ;advance time
.NoTimerAdvance:

   %if PlayInInterrupt
    sti                         ;enabled interrupts in case we take too long
    pusha
    call PlaySounds
    popa
   %endif

.NoPlay:
    and dword [PlayOptions],~PlayOptions.InInterrupt
    add dword [InterruptAccum],(1193182/30)<<16 ;18.2 / 30 ratio
    jnc .NoChain                ;take care of int ourselves
    pop eax
    pop es
    pop ds
    jmp far [cs:SetupTimers.OldHandler]
.NoChain:
    mov al,60h                  ;acknowledge interrupt (IRQ0)
    out 20h,al
    pop eax
    pop es
    pop ds
    iret


;
; Checks the command line parameters and sets options accordingly.
; Can return text with error.
; Does not directly end the program, but returns carry to tell the caller to
; end it.
;
; () (cf=error, edx=error message)
CheckStartParams:
    push ebp
    mov ebp,esp

    mov eax,0006h               ;get segment base address
    mov bx,[Program.PspSelector]
    int 31h
    mov esi,ecx
    shl esi,16
    mov si,dx                   ;CX:DX = 32-bit linear base address of segment


    movzx ecx,byte [esi+80h]
    add esi,81h
    add ecx,esi
    mov [.EndPtr],ecx
    call .GetNextParam          ;skip any initial space
.NextParam:
    jae .End                    ;cf=0
    push dword .NextParam
    push dword .CheckNextParam
    cmp byte [esi],'/'          ;is it an option?
    jne near .NonSwitchParam    ;if not, assume it might be a filename
    mov ebx,.List
    inc esi                     ;skip '/'
    call .MatchWord
    mov edx,Text.InvalidParameter
    jc .End
    jmp dword [.Jtbl+eax*4]
; (edx=error message ptr)
.Err:
    stc
; (edx=error message ptr if carry set)
.End:
    mov esp,ebp
    pop ebp
    ret

;
; Verifies the next parameter is separated properly by either a space, slash,
; null, or some other valid character.
; If all is okay, it will skip any separating space and return the new ptr.
; If there is an error in the parameter line, this routine will NOT return,
; but instead abort with an error message.
; (esi=param char ptr) (esi=new ptr)
.CheckNextParam:
    movzx eax,byte [esi]
    bt dword [.SepChars],eax
    mov edx,Text.InvalidParameter ;just in case invalid parameter
    jc .End                     ;cf=1

; skip any separating space and return the new ptr
; (esi=param char ptr) (esi=new ptr; *)
.GetNextParam:
    dec esi
.SkipSpace:
    inc esi
    cmp byte [esi],' '
    je .SkipSpace
    cmp esi,[.EndPtr]
    ;jae .End                    ;cf=0
    ret

;
; Searches through a string list for given text and returns matching index.
; Mainly used to match switch parameters, but can actually be used for any.
; (esi=search text, ebx=string list) (cf=error no match, eax=index)
.MatchWord:
    xor eax,eax                 ;start with first word in string list
    xor ecx,ecx
    ;cld
    jmp short .MwFirstCompare
.MwNextCompare:
    mov edx,ecx                 ;copy word length
    mov edi,ebx                 ;set to word in list for comparison
    rep cmpsb
    je .MwMatch                 ;word matched (cf=0)
    add esi,ecx                 ;compensate for change in esi after search
    add ebx,edx                 ;advance next string in list
    inc eax                     ;next word in list
    sub esi,edx
.MwFirstCompare:
    mov cl,[ebx]                ;get of length compare string
    inc ebx                     ;skip length byte to first character
    test ecx,ecx
    jnz .MwNextCompare          ;continue until word length = 0
    stc                         ;no words left, none matched
.MwMatch:                       ;match from above (cf=0)
    ret

;
; Determines length of word, up to next space, or within quotes.
; Mainly used to get filename, but can also be used for words following
; parameters.
; (esi=char ptr) (ecx=char length, ebx=ptr to following char; esi)
.GetWordLen:
    xor ecx,ecx
    cmp byte [esi],'"'
    mov ah,' '
    jne .GlsUnquoted
    inc esi
    mov ah,'"'
.GlsUnquoted:
    mov ebx,esi                 ;copy parameter char ptr
.GslNext:
    mov al,[ebx]                ;get char
    ;test al,al                 ;is null? for when in Wudebug
    ;jz .GslEnd
    cmp al,13                   ;for when normal command line start
    je .GslEnd
    cmp al,ah                   ;is space or ending quote
    je .GslEnd
    inc ebx
    inc ecx                     ;count another character in word
    jmp short .GslNext
.GslEnd:
    cmp al,'"'
    jne .GlsNoLastQuote
    inc ebx                     ;skip closing quote
.GlsNoLastQuote:
    ret

;
; Returns a number from the current parameter, adjusted char ptr to the
; character immediately after the last number.
; If the number is missing, this routine will NOT return, but instead abort
; with an error message.
; (esi=ptr to number)
; (eax=value, esi=first character after number)
.GetNumber:
    mov ebx,10                  ;default is decimal numbers
; (ebx=alternate radix)
.GetNumberOfRadix:
    call .GetNextParam          ;skip any initial space
    mov edx,Text.MissingParameter ;set message in case error
    jae near .Err               ;cf=0

    mov ecx,StringToNum.DefMaxLen
    call StringToNum.UsingLRadix
    mov edx,Text.MissingParameter ;set message in case error
    jz near .Err                ;no number so end
    ret

;
.NonSwitchParam:
    ;cld
    cmp byte [SpcFileName],0
    mov edx,Text.MultipleParameters
    jne near .Err
    call .GetWordLen
    mov edi,SpcFileName
    rep movsb
    mov byte [edi],0            ;put null on end of filename
    mov esi,ebx                 ;skip trailing " or space
    ret
.Help:
    call .CheckNextParam
    mov edx,Text.About
    jmp .Err
.SampleRate:
    call .GetNumber
    mov edx,Text.InvalidSampleRate  ;set message in case error
    cmp eax,Sound.MaxSampleRate
    ja near .Err
    cmp eax,Sound.MinSampleRate
    jb near .Err
    mov [Sound.SampleRate],eax
    mov [Sound.MixRate],eax
    ret
.MixRate:
    call .GetNumber
    mov edx,Text.InvalidMixingRate  ;set message in case error
    cmp eax,Sound.MaxMixRate
    ja near .Err
    cmp eax,Sound.MinMixRate
    jb near .Err
    mov [Sound.MixRate],eax
    ret
.Info:
    or dword [PlayOptions],PlayOptions.ShowInfo
    ret
.8bit:
    or byte [SoundFlags],SoundFlags.Force8bit
    ret
.16bit:
    or byte [SoundFlags],SoundFlags.Supports16bit
    ret
.Fm:
    ;or byte [SoundFlags],SoundFlags.Fm
    or dword [PlayOptions],PlayOptions.FmVoices
    ret
.DefaultPatch:
    call .GetNumber
    cmp eax,127
    mov edx,Text.InvalidPatch
    ja near .Err
    mov [DefaultGmPatch],eax
    ret
.As:
    ;or byte [SoundFlags],SoundFlags.Da
    or dword [PlayOptions],PlayOptions.AsVoices
    ret
.Sine:
    ;or byte [SoundFlags],SoundFlags.Da
    or dword [PlayOptions],PlayOptions.SineVoices
    ret
.Mpu:
    or byte [SoundFlags],SoundFlags.Mpu     ;audio output is disabled until
    ;and byte [SoundFlags],~SoundFlags.Da    ;I figure why my PC freezes
    or dword [PlayOptions],PlayOptions.MpuVoices
    ret
.Dls:
    ;or byte [SoundFlags],SoundFlags.Da
    or dword [PlayOptions],PlayOptions.DlsVoices
    ret
.NoFm:
    and byte [SoundFlags],~SoundFlags.Fm
    ret
.NoDa:
    and byte [SoundFlags],~SoundFlags.Da
    ret
.NoMpu:
    and byte [SoundFlags],~SoundFlags.Mpu
.NoGui:
    or dword [PlayOptions],PlayOptions.NoGui
    ret

;
section .data
align 4
.EndPtr:    dd 0
.SepChars:  ;characters that can separate parameters
    dd 11111111111111111101101111111110b  ;(control characters)
    dd 11111111111111110111111111111110b  ;?>=<;:9876543210/.-,+*)('&%$#"! 
    dd 11101111111111111111111111111111b  ;_^]\[ZYXWVUTSRQPONMLKJIHGFEDCBA@
    dd 11111111111111111111111111111111b  ;~}|{zyxwvutsrqponmlkjihgfedcba`
    dd 11111111111111111111111111111111b  ;
    dd 11111111111111111111111111111111b  ;
    dd 11111111111111111111111111111111b  ;
    dd 11111111111111111111111111111111b  ;

.Jtbl:      ;jump table for parameters
    dd .Help
    dd .Help
    dd .SampleRate
    dd .Mpu
    dd .MixRate
    dd .Info
    dd .8bit
    dd .16bit
    dd .Fm
    dd .As
    dd .Sine
    dd .Dls
    dd .NoFm
    dd .NoDa
    dd .NoMpu
    dd .NoGui
    dd .DefaultPatch
.List:      ;parameter word strings
    ;first byte is length of string
    ;following bytes are characters of parameter string (1-255)
    ;last length of list is zero
    db 1,"h"
    db 1,"?"
    db 1,"r"
    db 3,"mpu"
    db 1,"m"
    db 4,"info"
    db 1,"8"
    db 2,"16"
    db 2,"fm"
    db 2,"as"
    db 4,"sine"
    db 3,"dls"
    db 4,"nofm"
    db 4,"noda"
    db 5,"nompu"
    db 5,"nogui"
    db 2,"dp"
    db 0
section .text

;
; (esi=source)
; (ecx=length, zf=zero length; esi,edx)
GetStringLength:
    mov eax,1024            ;maximum length of characters
; (al=character to search for, eax=maximum length of characters to search)
.UntilChar:
    mov ecx,eax             ;make a copy of max length for later
    mov edi,esi             ;copy source for string length search
    cld                     ;as always, look forward
    repne scasb             ;search for the end, until character is found
    ;neg ecx                ;;get negative count
    ;dec ecx                ;;minus the character at the end
    not ecx                 ;negate count and subtract character at the end
    add ecx,eax             ;get length (conveniently sets zf)
    ret


;
; (esi=source of text)
StatusMessage:
    ;test byte [Gui.Status],Gui.StatusActive
    ;jz .DosPrint
    ;not done
    ;use gui instead of DOS to print msg
    ;ret
.DosPrint:
    mov edx,esi
    jmp WriteString.WithEol
    ;ret


;
; Turns a 32bit number into a decimal (or other) string, writing it to edi.
; By default, it converts a number to a decimal string, maximum of ten
; characters, stored in NumToString.Buffer. To change where the string is
; stored, the length, or radix, these variables can be passed to a different
; entry point. If an alternate destination is supplied, the default is a
; decimal number up to ten characters. If a length is supplied, so must the
; destination; and if the radix, so must both destination and length be given.
;
; (eax=number, edi=destination, ?ecx=maximum length, ?ebx=radix)
; (ecx=offset of first significant digit, ebx=radix used; esi)
NumToString:
    mov edi,.Buffer
;(edi=destination)
.UsingDest:
    ;mov ecx,[.MaxLen]       ;default maximum of ten characters, since the largest 32bit number is 4 gigabytes
    mov ecx,.DefMaxLen
;(edi=destination, ecx=number of digits)
.UsingDLen:
	mov ebx,10              ;base of the decimal system
;(edi=destination, ecx=number of digits, ebx=radix)
.UsingDLRadix:              ;for hexadecimal and binary (even octal)
	xor edx,edx             ;set top 32 bits of quotient to zero
    lea edi,[edi+ecx-1]     ;start from rightmost character in number
.NextChar:
	div ebx                 ;divide number by the decimal base
    mov dl,[.NumberTable+edx] ;get ASCII character from table
    ;add dl,'0'             ;make remainder into an ASCII character
    mov [edi],dl            ;output result
	dec edi                 ;move backwards one character
	test eax,eax            ;see if we are done with the number
	jz .FillInBlanks        ;nothing but zeroes left
	xor edx,edx             ;set edx to zero again for next division
	dec ecx                 ;one less character to output
	jnz .NextChar
	ret

.FillInBlanks:
    mov al,[.FillChar]      ;fill in with spaces, zeroes, asterisks
	dec ecx                 ;one less than current count
	mov edx,ecx
    std                     ;move backwards
    rep stosb               ;for number of characters remaining
    mov ecx,edx             ;return offset of first digit
	ret

section .data
.DefMaxLen      equ 10
;.MaxLen:        dd .DefMaxLen
.FillChar:      db ' '
.NumberTable:   db "0123456789ABCDEF"
section .bss
.Buffer:        resb .DefMaxLen
section .text


;
; String to Number
; (esi=text source, ?ebx=radix, ?ecx=length of string)
; (eax=value, esi=character immediately after, zf=no number)
;
; Turns a string representation of a number into a 32bit unsigned number.
; Ends at a non-numeric digit, including puncuation, extended characters,
; null, or any other control character. Returns zero for a an empty string.
;
StringToNum:
    ;set default returned number to zero
    ;start at first number
    ;do until number (<0 >9 <A >Z)
    ;  multiply by radix and add to value
    ;loop
    ;return value and string length
    mov ecx,.DefMaxLen
.UsingLength:
    mov ebx,10              ;base of the decimal system
.UsingLRadix:               ;for hexadecimal and binary (even octal)
    xor edx,edx             ;set top 32 bits of digit place to zero
    mov edi,ecx             ;copy length
    xor eax,eax             ;set return value to zero
.NextChar:
    mov dl,[esi]            ;get digit
    sub dl,48               ;"0"=48->0  "9"=57->9
    jc .End                 ;character is less than '0'
    cmp dl,10
    jb .AddPlaceValue       ;digit is 0-9
    and dl,~32              ;make uppercase by turning off fifth bit
    cmp dl,'A'-48
    jb .End                 ;character is less than 'A'
    cmp dl,'F'-48
    ja .End                 ;character is greater than 'F' (15)
    sub dl,7                ;(65-48)+10  "A"=65->10  "F"=86->15
.AddPlaceValue:
    cmp dl,bl
    jae .End
    imul eax,ebx            ;multiply existing number by radix
    add eax,edx             ;add new digit
    inc esi                 ;move forwards one character
    dec cl                  ;one less character to check
    jnz .NextChar
.End:
    sub edi,ecx             ;set zero flag accordingly
    ;mov ecx,edi
    ret

.DefMaxLen  equ 10          ;default maximum since the largest 32bit is 4gb

;
; (eax=midi patch number)
; (esi=ptr to name, ecx=char length)
GetMidiPatchName:
    ;and eax,127
    mov esi,Text.InstrumentList
    xor ecx,ecx
.Next:
    add esi,ecx
    mov cl,[esi]
    inc esi
    dec al
    jns .Next
    ret


;
; (esi=asciiz text ptr) ()
SetWindowsTitle:
    call GetStringLength        ; (esi=source) (ecx=length)

    xor ebx,ebx
    mov [Regs.eax],dword 168Eh  ;set/get Window attribute
    mov [Regs.ecx],ecx          ;set string length (seems unnecessary)
    mov [Regs.ebx],ebx
    mov [Regs.edx],ebx          ;set window title
    mov [Regs.edi],ebx          ;zero string ptr
    mov [Regs.es],word 0BA00h
    ; just use untouched text memory for temp storage
    ; rather than hassle with allocating low DOS memory
    mov edi,0BA000h
    inc ecx                     ;include null
    cld
    rep movsb

    mov eax,0300h               ;simulate real mode interrupt
    mov ebx,2Fh                 ;DOS multiplex interrupt
    ;xor ecx,ecx                ;no stack copy
    mov edi,Regs
    int 31h

    ;bt dword [Regs.Flags],0    ;set carry if error
    ret

%if 0
;
; Simply wipes ranges of memory using a ptr/length list.
;
ZeroMemory:
.Next:
    cld
    xor eax,eax
.Next:
    mov edi,[esi]
    mov ecx,[esi+4]             ;must be in dwords (not bytes!)
    test edi,edi
    jz .End
    rep stosd
    ;mov ebx,ecx
    ;and ecx,3
    ;rep stosb
    add esi,byte 8
    jmp short .Next
    ret
%endif


%if 0
;
; ()() zero based selector is stored in AbsoluteSelector
GetAbsoluteSelector:
    xor eax,eax                 ;DPMI function 0, allocate selector
    mov ecx,1                   ;only one selector
	int 31h
	jc .ErrorAllocating	;uh-oh
    mov [AbsoluteSelector],eax  ;store returned selector
    xor edx,edx                 ;set base address to zero
    mov ebx,eax                 ;copy selector
    xor ecx,ecx                 ;put upper 16 bits into ecx (cx:dx)
    mov eax,7                   ;DPMI function 7, set base address
	int 31h
    jc .ErrorSetting            ;could not set base
	xor edx,edx
    mov ecx,(1048576*2)>>16     ;put upper 16 bits into ecx (cx:dx)
    mov eax,8                   ;DPMI function 8, set segment limit
	int 31h
    jc .ErrorSetting            ;could not set limit
	ret

.ErrorSetting:
	mov ebx,[AbsoluteSelector]
	mov eax,1		;DPMI function 1, free selector
	int 31h			;ebx should still contain our selector
	stc
.ErrorAllocating:
	mov edx,Text.ErrorGettingSelector
	ret

align 4
AbsoluteSelector:       dd 0
%endif


;

%include "spcemu.asm"
%include "dspemu.asm"
%include "dspsim.asm"
%include "dspmisc.asm"

%include "disasm.asm"

%include "file.asm"
%include "settings.asm"

%include "midi.asm"
%include "sound.asm"

%include "gfx.asm"
%include "spcview.asm"

%if 0
DebugMode   equ 0
%define UseGui
%define UseGuiGfx
%define UseGuiCursor
%define UseGuiFonts
%define UseWindowCode
%define UseTitleBarCode
%define UseTextPromptCode
%define UseLabelCode
%define UseImageCode
%define UseBorderCode
%define UseScrollBarCode
%define UseButtonCode
%include "..\gui\guidefs.asm"
%include "..\gui\gfxdefs.asm"
%include "gui.asm"
%include "..\gui\guicode.asm"
%include "..\gui\guiobjs.asm"
%include "..\gui\gfx.asm"
%endif

%if ShowDebugInfo
%include "debug.asm"
%endif

section .data
;
Text:
.Title:             db "Spc2Midi ",ProgVersion," - (incomplete) DSP to MIDI converter. (c)2001 PeekinSot",13,10
                    db 10,0
.WinTitle:          db "Spc2Midi ",ProgVersion,0
.About:             db "Now that the SPC-700 emulation core is complete, I can start working on MIDI",13,10
                    db "exporting, saving sample settings, and of course, a GUI! Currently, it does",13,10
                    db "little more than load an SPC state (ZST,ZMV,SPC), emulate the processor,",13,10
                    db "sound the notes, and graphically show them.",13,10
                    db 10
                    db "Usage:  Spc2Midi [/options] filename.ext",13,10
                    db 10
                    db "Options:",13,10
                    db "  as    - audio samples (default)",13,10
                    db "  fm    - play FM synthesis",13,10
                    db "  sine  - sine wave synthesis",13,10
                   ;db "  dls   - (just an idea for now)",13,10
                    db "  mpu   - MPU output (!untested!)",13,10
                    db "  nofm  - disable FM initialization",13,10
                    db "  noda  - disable wave audio",13,10
                    db "  nompu - disable MPU initialization",13,10
                    db "  r#    - sample rate 5000-64000  (default=32000)",13,10
                    db "  m#    - mixing rate 2000-200000 (default=32000)",13,10
                    db "  dp#   - default GM patch (0-127)",13,10
                    db "  8     - force 8bit on 16bit sound cards (default=16)",13,10
                    db "  16    - force 16bit on cards <= DSP3",13,10
                    db "  info  - sample table information",13,10
                    db "  nogui - play with only the black screen you see now",13,10
                    db "  h ?   - the info you are reading now",13,10
                    db 0
.UnverifiedOpcode:  db "##    @#### <########> - Unverified opcode. Emulation may procede incorrectly.",13,10,0
.DspWrite:          db "##:## @#### <########> - DSP written to.",13,10,0
.DspRead:           db "##:## @#### <########> - DSP read from.",13,10,0
.DspWriteIllegal:   db "##:## @#### <########> - Write to invalid DSP register.",13,10,0
.RegWriteIllegal:   db "##:## @#### <########> - Write to read-only register.",13,10,0
.RegReadIllegal:    db "##:## @#### <########> - Read from write-only register.",13,10,0
.MissingParameter:  db "! Missing parameter value",13,10,0
.InvalidParameter:  db "! Invalid parameter",13,10,0
.MultipleParameters:db "! More than one filename was given",13,10,0
.InvalidSampleRate: db "! Invalid sample rate (5000-64000)",13,10,0
.InvalidMixingRate: db "! Invalid mixing rate (2000-200000)",13,10,0
.InvalidPatch:      db "! Invalid GM patch (0-127)",13,10,0
.FileTypeBad:       db "Unrecognized filetype. Spc2Midi can only load SPC captures and savestates.",13,10
                    db "Valid filetypes are SPC, ZST, or ZMV.",13,10,0
.FileOpenError:     db "Could not open the given file. Check the spelling and path.",13,10,0
.SpcReadError:      db "File does not contain SPC data or is corrupt!",13,10,0
.DspOutputError:    db "Could not open the DSP output file!",13,10,0
.LoadingFile:       db "Loading CPU and DSP state.",0
.InitializingWave:  db "Initializing wave audio",0
.InitializingFm:    db "Initializing FM synthesis",0
.InitializingMpu:   db "Initializing MPU (untested!)",0
.NoBlasterVars:     db "! No 'BLASTER' variable set",0
.SbNotReset:        db "! Sound card would not reset",0
.FmNotReset:        db "! FM either busy or unsupported",0
.MpuNotReset:       db "! MPU either busy or unsupported",0
.StartingSound:     db "Starting sound transfer",0
.StoppingSound:     db "Stopping sound transfer",0
.ReleasingWave:     db "Releasing wave audio resources",0
.ReleasingFm:       db "Silencing FM voices",0
.ReleasingMpu:      db "Stopping MPU output",0
.ValidatingSamples: db "Checking valid BRR samples",0
.BufferingSamples:  db "Buffering BRR sound samples",0
.IdentifyingSamples:db "Identifying unique samples",0
;.LoadingSettings:  db "Loading settings from xxxxxxxx.xxx",0
.Ready:             db "Ready for emulation. Press a key...",13,10
                    db 10
                    db "Muting:",13,10
                    db "  1-8    Mute/Enable",13,10
                    db "  0      Mute all",13,10
                    db "  9      Enable all",13,10
                    db 10
                    db "Play Control:",13,10
                    db "  Space  Pause",13,10
                    db "  Home   Restart",13,10
                    db "  End    End of emulation",13,10
                    db "  BkSpc  Reverse play",13,10
                    db "  - +    Change tempo",13,10
                    db "  <- ->  Seek one second",13,10
                    db 10
                    db "Sound Output:",13,10
                    db "  a      Toggle audio samples",13,10
                    db "  s      Toggle sine wave synth",13,10
                   ;db "  d      (someday) toggle DLS",13,10
                    db "  f      Toggle FM",13,10
                    db "  g      Toggle GM MPU401",13,10
                    db 10
                    db "Other:",13,10
                    db "  ",24," ",25,"    Change sample",13,10
                    db "  Esc    Quit",13,10
                    db 0
.SampleInfoTitles:  db "Sample    Offset   Valid   Length    Checksum",13,10,0
.SampleInfo:        db "###    ##### #####  ##  ##### #####  ########  $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$",13,10,0
.CurTime:           db "###:##.### / ###:##.###",13,0
.InstrumentList:
; PIANO
db 14,"Acoustic Grand"
db 15,"Bright Acoustic"
db 14,"Electric Grand"
db 10,"Honky-Tonk"
db 16,"Electric Piano 1"
db 16,"Electric Piano 2"
db 11,"Harpsichord"
db  8,"Clavinet"
; CHROMATIC PERCUSSION
db  7,"Celesta"
db 12,"Glockenspiel"
db 9,"Music Box"
db 10,"Vibraphone"
db  7,"Marimba"
db  9,"Xylophone"
db 13,"Tubular Bells"
db  8,"Dulcimer"
; ORGAN
db 13,"Drawbar Organ"
db 16,"Percussive Organ"
db 10,"Rock Organ"
db 12,"Church Organ"
db 10,"Reed Organ"
db  9,"Accordian"
db  9,"Harmonica"
db 15,"Tango Accordian"
; GUITAR
db 19,"Nylon String Guitar"
db 19,"Steel String Guitar"
db 20,"Electric Jazz Guitar"
db 21,"Electric Clean Guitar"
db 21,"Electric Muted Guitar"
db 17,"Overdriven Guitar"
db 17,"Distortion Guitar"
db 16,"Guitar Harmonics"
; BASS
db 13,"Acoustic Bass"
db 22,"Electric Bass (finger)"
db 20,"Electric Bass (pick)"
db 13,"Fretless Bass"
db 11,"Slap Bass 1"
db 11,"Slap Bass 2"
db 12,"Synth Bass 1"
db 12,"Synth Bass 2"
; SOLO STRINGS
db  6,"Violin"
db  5,"Viola"
db  5,"Cello"
db 10,"Contrabass"
db 15,"Tremolo Strings"
db 17,"Pizzicato Strings"
db 18,"Orchestral Strings"
db  7,"Timpani"
; ENSEMBLE
db 17,"String Ensemble 1"
db 17,"String Ensemble 2"
db 14,"SynthStrings 1"
db 14,"SynthStrings 2"
db 10,"Choir Aahs"
db 10,"Voice Oohs"
db 11,"Synth Voice"
db 13,"Orchestra Hit"
; BRASS
db  7,"Trumpet"
db  8,"Trombone"
db  4,"Tuba"
db 13,"Muted Trumpet"
db 11,"French Horn"
db 13,"Brass Section"
db 12,"SynthBrass 1"
db 12,"SynthBrass 2"
; REED
db 11,"Soprano Sax"
db  8,"Alto Sax"
db  9,"Tenor Sax"
db 12,"Baritone Sax"
db  4,"Oboe"
db 12,"English Horn"
db  7,"Bassoon"
db  8,"Clarinet"
; PIPE
db  7,"Piccolo"
db  5,"Flute"
db  8,"Recorder"
db  9,"Pan Flute"
db 12,"Blown Bottle"
db 10,"Skakuhachi"
db  7,"Whistle"
db  7,"Ocarina"
; SYNTH LEAD
db 15,"Lead 1 (square)"
db 17,"Lead 2 (sawtooth)"
db 17,"Lead 3 (calliope)"
db 14,"Lead 4 (chiff)"
db 16,"Lead 5 (charang)"
db 14,"Lead 6 (voice)"
db 15,"Lead 7 (fifths)"
db 18,"Lead 8 (bass+lead)"
; SYNTH PAD
db 15,"Pad 1 (new age)"
db 12,"Pad 2 (warm)"
db 17,"Pad 3 (polysynth)"
db 13,"Pad 4 (choir)"
db 13,"Pad 5 (bowed)"
db 16,"Pad 6 (metallic)"
db 12,"Pad 7 (halo)"
db 13,"Pad 8 (sweep)"
; SYNTH EFFECTS
db 11,"FX 1 (rain)"
db 17,"FX 2 (soundtrack)"
db 14,"FX 3 (crystal)"
db 17,"FX 4 (atmosphere)"
db 17,"FX 5 (brightness)"
db 14,"FX 6 (goblins)"
db 13,"FX 7 (echoes)"
db 13,"FX 8 (sci-fi)"
; ETHNIC
db  5,"Sitar"
db  5,"Banjo"
db  8,"Shamisen"
db  4,"Koto"
db  7,"Kalimba"
db  7,"Bagpipe"
db  6,"Fiddle"
db  6,"Shanai"
; PERCUSSIVE
db 11,"Tinkle Bell"
db  5,"Agogo"
db 11,"Steel Drums"
db  9,"Woodblock"
db 10,"Taiko Drum"
db 11,"Melodic Tom"
db 10,"Synth Drum"
db 14,"Reverse Cymbal"
; SOUND EFFECTS
db 17,"Guitar Fret Noise"
db 12,"Breath Noise"
db  8,"Seashore"
db 10,"Bird Tweet"
db 14,"Telephone Ring"
db 10,"Helicopter"
db  8,"Applause"
db  7,"Gunshot"


SoundDumpHeader:    db "RIFF"
                    dd 4+24+8+SamplesBufferSize
                    db "WAVEfmt "
                    dd 16
                    dw 1,1          ;PCM/Mono
                    dd 11025,22050  ;much lower than actual rate (32khz)
                    ;dd 8000,16000   ;much lower than actual rate (32khz)
                    dw 1,16         ;block alignment/16bit audio
                    db "data"
                    dd SamplesBufferSize
SoundDumpHeader.Size equ $-SoundDumpHeader
SoundDumpFileName:  db "out.wav",0
